/*
** $Id: agnhlps.c $
** C Helper routines
** See Copyright Notice in agena.h
** initiated Alexander Walz, July 20, 2007
*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <fcntl.h>

#ifdef _WIN32
#include <sys/locking.h>
#include <windows.h>  /* free memory */
#include <math.h>     /* for pow */
#endif

#include <sys/stat.h>

#if defined(__unix__) || defined(__APPLE__)  /* for getch */
#include <termios.h>
#include <unistd.h>
#endif

#if defined(__OS2__)
#include <unistd.h>
#endif

#include <limits.h>    /* for PATH_MAX length */

#include "agnhlps.h"
#include "agncfg.h"  /* read in endianness */
#include "agncmpt.h" /* for O_BINARY */
#include "agnconf.h"

#define agnhlps_c
#define AGENA_LIB

#define BASE_MOVELEN  16384L     /* number of bytes in index section to be moved in a write or delete operation */
#define BASE_OFFSET     256L     /* first position of an index to a key */
#define MAXNRECORDS     104L     /* position of Maximum Records allowed */
#define COMMENT         129L     /* position of comment */

# define uchar(c) ((unsigned char)(c))

/* swap() and quicksort() are used in ADS package */
void swap (off64_t v[], int i, int j) {
  int temp;
  temp = v[i];
  v[i] = v[j];
  v[j] = temp;
}


void tools_quicksort (off64_t v[], int left, int right) {
  int i, last;
  if (left >= right) return;
  swap(v, left, (left+right)/2);
  last = left;
  for (i=left+1; i<=right; i++)
    if (v[i]<v[left]) {
      swap(v, ++last, i);
    }
  swap(v, left, last);
  tools_quicksort(v, left, last-1);
  tools_quicksort(v, last+1, right);
}


/* Dr. F.H.Toor's database C functions */

int my_open (const char *file) {
  int hnd;
  hnd = open(file, O_BINARY | O_RDWR);
  if (hnd < 3)
    return -1;
  else
    return(hnd);
}


/* read-only open */

int my_roopen (const char *file) {
  int hnd;
  hnd = open(file, O_BINARY | O_RDONLY); /*|_S_IWRITE); */
  if (hnd < 3) {
    return -1;
  } else
    return(hnd);
}


int my_create (const char *file) {
  int hnd;
  hnd = open(file, O_BINARY | O_RDWR | O_CREAT | O_TRUNC, S_IREAD | S_IWRITE);  /* GCC 4.3.2 needs a 3rd arg */
  if (hnd < 3) {
    /* fprintf(stderr, "create error\n"); */
    return -1;
  } else
    return(hnd);
}


int my_close (int hnd) {
  return close(hnd);
}


int my_seek (int hnd, int32_t pos) {
  if (lseek64(hnd, pos, SEEK_SET) == -1) {
    return(-1);
  } else
    return(1);
}


off_t my_lof (int hnd) {
  off_t size;
  size = lseek64(hnd, 0, SEEK_END);
  if (size == -1)
    fprintf(stderr, "Agena IO subsystem: LOF error\n");
  return(size);
}


void my_read (int hnd, void *data, size_t size) {
  if (read(hnd, data, size) != size) {
    fprintf(stderr, "Agena IO subsystem: read error\n");
  }
}


size_t sec_read (int hnd, void *data, size_t size) {
  return (read(hnd, data, size) == size);
}


size_t sec_readl (int hnd, ssize_t *success) {
  int32_t data;
  *success = sec_read(hnd, &data, sizeof(int32_t));
#if BYTE_ORDER != BIG_ENDIAN
  tools_swapint32_t(&data);
#endif
  return(data);
}


/* longs are always stored in Big Endian notation */

int32_t my_readl (int hnd) {
  int32_t data;
  my_read(hnd, &data, sizeof(int32_t));
#if BYTE_ORDER != BIG_ENDIAN
  tools_swapint32_t(&data);
#endif
  return(data);
}


char my_readc (int hnd) {
  char data;
  my_read(hnd, &data, sizeof(char));
  return(data);
}


void my_write (int hnd, void *data, size_t size) {
  if (write(hnd, data, size) != size)
    fprintf(stderr, "Agena IO subsystem: write error\n");
}


/* longs are always stored in Big Endian notation */

void my_writel (int hnd, int32_t data) {
#if BYTE_ORDER != BIG_ENDIAN
  tools_swapint32_t(&data);
#endif
  my_write(hnd, &data, sizeof(int32_t));
}


void my_writec (int hnd, char data) {
  my_write(hnd, &data, sizeof(char));
}


/* in UNIX and Windows, if size == 0 then lock entire file */
int my_lock (int hnd, off64_t start, off64_t size) {
  #ifdef _WIN32
  /* see: http://msdn.microsoft.com/en-us/library/8054ew2f.aspx, topic `_locking` */
  int r;
  off64_t oldpos, cursor;
  if (size == 0) size = (off64_t)pow(2, 63);  /* = lock 2^63 bytes (cannot be more in Windows) */
  /* reset file cursor to `start' */
  oldpos = my_fpos(hnd);
  if (oldpos == -1) return -1;
  cursor = lseek64(hnd, start, SEEK_SET);
  if (cursor == -1) return -1;
  /* lock the file */
  r = _locking(hnd, _LK_NBLCK, size);
  /* set file cursor to original file position */
  cursor = lseek64(hnd, oldpos, SEEK_SET);
  if (cursor == -1 || r != 0) return -1;
  #elif defined(__unix__) || defined(__APPLE__)
  struct flock file;
  file.l_type = F_WRLCK | F_RDLCK;   /* exclusive read and write lock */
  file.l_start = (off_t)start;   /* locking from file position `start` */
  file.l_whence = SEEK_SET;      /* offset is relative to the beginning of the file */
  file.l_len = (size == 0) ? 0L : (off_t)size;  /* lock size bytes (size == 0 query: better sure than sorry) */
  if (fcntl(hnd, F_SETLK, &file) == -1)
    return -1;
  #endif
  return 0;
}


int my_unlock (int hnd, off64_t start, off64_t size) {
  #ifdef _WIN32
  /* see: http://msdn.microsoft.com/en-us/library/8054ew2f.aspx, topic `_locking` */
  int r;
  off64_t oldpos, cursor;
  if (size == 0) size = (off64_t)pow(2, 63);  /* = lock 2^63 bytes (cannot be more in Windows) */
  /* reset file cursor to `start' */
  oldpos = my_fpos(hnd);
  if (oldpos == -1) return -1;
  cursor = lseek64(hnd, start, SEEK_SET);
  if (cursor == -1) return -1;
  /* lock the file */
  r = _locking(hnd, LK_UNLCK, size);
  /* set file cursor to original file position */
  cursor = lseek64(hnd, oldpos, SEEK_SET);
  if (cursor == -1 || r != 0) return -1;
  #elif defined(__unix__) || defined(__APPLE__)
  struct flock file;
  file.l_type = F_WRLCK | F_RDLCK;
  file.l_whence = SEEK_SET;
  file.l_start = (off_t)start;
  file.l_len = (size == 0) ? 0L : (off_t)size;  /* unlock size bytes (size == 0 query: better sure than sorry) */
  if (fcntl(hnd, F_UNLCK, &file) == -1)
    return -1;
  #endif
  return 0;
}


/* my_move(hnd, low*4L+24L, low*4L+28L, cnt*4L+24L); */
/* fpos: from position, tpos: to position, size */
void my_move (int hnd, int32_t fpos, int32_t tpos, int32_t size) {
  int32_t segment;
  char buff[BASE_MOVELEN];
  segment = size-fpos;
  if (segment < 1)
    return;
  if (fpos < 0 || tpos < 0)
    fprintf(stderr, "Agena IO subsystem: move error\n");
  if (fpos > tpos) {
    while (segment >= BASE_MOVELEN) {
      my_seek(hnd, fpos);
      my_read(hnd, buff, BASE_MOVELEN);
      my_seek(hnd, tpos);
      my_write(hnd, buff, BASE_MOVELEN);
      fpos+=BASE_MOVELEN;
      tpos+=BASE_MOVELEN;
      segment-=BASE_MOVELEN;
    }
    if (segment > 0) {
      my_seek(hnd, fpos);
      my_read(hnd, buff, segment);
      my_seek(hnd, tpos);
      my_write(hnd, buff, segment);
    }
  }
  if (fpos < tpos) {
    while (segment >= BASE_MOVELEN) {
      my_seek(hnd, fpos+segment-BASE_MOVELEN);
      my_read(hnd, buff, BASE_MOVELEN);
      my_seek(hnd, tpos+segment-BASE_MOVELEN);
      my_write(hnd, buff, BASE_MOVELEN);
      segment-=BASE_MOVELEN;
    }
    if (segment > 0) {
      my_seek(hnd, fpos);
      my_read(hnd, buff, segment);
      my_seek(hnd, tpos);
      my_write(hnd, buff, segment);
    }
  }
}


/* return current file position, written by Alexander Walz on June 28, 2007; extended
   0.32.5 */

off64_t my_fpos (int hnd) {
  off64_t result = lseek64(hnd, 0, SEEK_CUR);
  if (result == -1) {
    fprintf(stderr, "Agena IO subsystem: error while determinig file position\n");
    return(-1);
  } else
  return(result);
}


/* my_expand: core functionality for expanding a database by a) reshifting all records, thus
   b) adding more index entries.
   Assumes that the file is open and locked when envoked.
   Restores the file cursor to the position before it was called.

   written by Alexander Walz on June 28, 2007

   Arguments:
      hnd   - the file handler
      mrc   - maximum number of records currently allowed
      cnt   - current number of actual records
      count - number of records to be added */

void my_expand (int hnd, int mrc, int cnt, int count) {
  int32_t dsbegin, dsend, offset, j, index, bufsize, cfpos, commentpos, commentlen, newindex;
  int i;
  cfpos = my_fpos(hnd);           /* save the current file position for later restoration */
  my_seek(hnd, COMMENT);
  commentpos = my_readl(hnd);
  commentlen = 0;
  if (commentpos != 0) {
    my_seek(hnd, commentpos);
    commentlen = my_readl(hnd)+4L;
  }
  dsbegin = mrc*4L+BASE_OFFSET;   /* start of current dataset section */
  dsend = my_lof(hnd)-1;          /* end of current dataset section */
  /* write 0L at end of file count times */
  my_seek(hnd, dsend+1);
  for (i=1; i<=count; i++) {
    my_writel(hnd, 0L);          /* write zeros */
  }
  offset = count * 4L;
  char buffer[offset];
  /* beginning with the last data set move data sets to the new file end */
  bufsize = offset;
  j = dsend+1;
  while (j > dsbegin) {
    if ((j-dsbegin) >= offset) {
      bufsize = offset;
      my_seek(hnd, j-bufsize);
      my_read(hnd, buffer, bufsize);
      my_seek(hnd, j);
      my_write(hnd, buffer, bufsize); }
    else {  /* last few bytes need special treatment */
      my_seek(hnd, dsbegin);
      my_read(hnd, buffer, j-dsbegin);
      my_seek(hnd, j+bufsize-(j-dsbegin));
      my_write(hnd, buffer, j-dsbegin);
    }
    j -= bufsize;
  }
  /* update indices */
  for (j=0; j<cnt; j++) {
    my_seek(hnd, j*4L+BASE_OFFSET);
    index = my_readl(hnd);
    my_seek(hnd, j*4L+BASE_OFFSET);
    newindex = (index > commentpos) ? index+offset+commentlen : index+offset;
    my_writel(hnd, newindex);
  }
  /* add new indices to end of index section */
  for (j=0; j<count; j++) {
    my_writel(hnd, 0L);
  }
  /* update maximum number of entries field */
  my_seek(hnd, MAXNRECORDS);
  my_writel(hnd, mrc+count);
  /* update position of comment if present */
  if (commentpos != 0) {
    my_seek(hnd, COMMENT);
    my_writel(hnd, commentpos+offset);
  }
  my_seek(hnd, cfpos);  /* reset cursor to position before the function was called */
}


#if defined(__unix__) || defined(__APPLE__)
static struct termios new_io;
static struct termios old_io;

/* Taken from `C von A bis Z` by Jrgen Wolf, 2nd Ed., pp. 452 */
/* Funktion schaltet das Terminal in den cbreak-Modus:         */
/* Kontrollflag ECHO und ICANON auf 0 setzen                   */
/* Steuerzeichen: Leseoperation liefert 1 Byte VMIN=1 VTIME=1  */
int cbreak(int fd) {
  /*Sichern unseres Terminals*/
  if((tcgetattr(fd, &old_io)) == -1)
    return -1;
  new_io = old_io;
  /* Wir verndern jetzt die Flags fr den cbreak-Modus */
  new_io.c_lflag = new_io.c_lflag & ~(ECHO|ICANON);
  new_io.c_cc[VMIN] = 1;
  new_io.c_cc[VTIME]= 0;
  /* Jetzt setzen wir den cbreak-Modus */
  if((tcsetattr(fd, TCSAFLUSH, &new_io)) == -1)
    return -1;
  return 1;
}


int getch(void) {
  int c;
  if(cbreak(STDIN_FILENO) == -1) {
    printf("Agena IO subsystem: error in function cbreak (called by getch).\n");
    exit(EXIT_FAILURE);
  }
  c = getchar();
  /* Alten Terminal-Modus wiederherstellen */
  tcsetattr(STDIN_FILENO, TCSANOW, &old_io);
  return c;
}
#endif

/* Copyright (c) 2008    Johnathon

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to
deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
sell copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
IN THE SOFTWARE.

Source: http://code.google.com/p/linux-kbhit/linux_kbhit.tar.gz

Return: -1 = key not pressed

*/

#if defined(__unix__)

int kbhit (void) {  /* does not work on Mac */
  struct termios oldt, newt;
  int ch;
  tcgetattr( STDIN_FILENO, &oldt );
  newt = oldt;
  newt.c_cc[VMIN] = 0;
  newt.c_cc[VTIME] = 1;
  newt.c_lflag &= ~( ICANON | ECHO );
  tcsetattr( STDIN_FILENO, TCSANOW, &newt );
  ch = getchar();
  tcsetattr( STDIN_FILENO, TCSANOW, &oldt );
  return ch;
}

#endif


const char *tools_cwd (void) {
  char buffer[PATH_MAX];
  return (const char *) getcwd(buffer, PATH_MAX);
}


/* Determination of endianness; taken from Jrgen Wolf, `C von A bis Z`, 2. Auflage, Galileo Computing */

char tools_endian (void) {
  unsigned int Word = 0x22CCDDEE;
  unsigned char *Byte;
  Byte = (unsigned char *) &Word;
  if (Byte[0] == ((Word >> 0) & 0xFF))
    return 0;  /* little endian */
  else if (Byte[0] == ((Word >> 24) & 0xFF))
    return 1;  /* big endian */
  else
    return -1;
}


double tools_intpow (double x, int n) {
  double r = 1.0;
  if (n < 0) {
    if (x == 0) return AGN_NAN;  /* 0.32.4 patch */
    x = 1.0 / x;
    n = -n;
  }
  do {
    if (n & 1) r *= x;
    n >>= 1;
    x *= x;
  } while (n);
  return r;
}


/* is number finite ? */
int tools_isfinite (const double a) {
  const double b = a - a;
  int r = (b == b);
  return r;
}


/* is number NaN ? */
int tools_isnan (const double x) {
  int r = (x != x);
  return r;
}


/* define NaN */
double tools_nan (void) {
  return (0.0/0.0);
}


/* code taken from http://www.dmh2000.com/cpp/dswap.shtml */

#if BYTE_ORDER != BIG_ENDIAN
uint64_t tools_swapdouble (double d) {
  uint64_t a;
  unsigned char *dst = (unsigned char *)&a;
  unsigned char *src = (unsigned char *)&d;
  dst[0] = src[7];
  dst[1] = src[6];
  dst[2] = src[5];
  dst[3] = src[4];
  dst[4] = src[3];
  dst[5] = src[2];
  dst[6] = src[1];
  dst[7] = src[0];
  return a;
}

/* unswap using char pointers */
double tools_unswapdouble (uint64_t a) {
  double d;
  unsigned char *src = (unsigned char *)&a;
  unsigned char *dst = (unsigned char *)&d;
  dst[0] = src[7];
  dst[1] = src[6];
  dst[2] = src[5];
  dst[3] = src[4];
  dst[4] = src[3];
  dst[5] = src[2];
  dst[6] = src[1];
  dst[7] = src[0];
  return d;
}
#else
/* BIG ENDIAN */
uint64_t tools_swapdouble (double d) {
  uint64_t a;
  unsigned char *dst = (unsigned char *)&a;
  unsigned char *src = (unsigned char *)&d;
  dst[0] = src[0];
  dst[1] = src[1];
  dst[2] = src[2];
  dst[3] = src[3];
  dst[4] = src[4];
  dst[5] = src[5];
  dst[6] = src[6];
  dst[7] = src[7];
  return a;
}

/* unswap using char pointers */
double tools_unswapdouble (uint64_t a) {
  double d;
  unsigned char *src = (unsigned char *)&a;
  unsigned char *dst = (unsigned char *)&d;
  dst[0] = src[0];
  dst[1] = src[1];
  dst[2] = src[2];
  dst[3] = src[3];
  dst[4] = src[4];
  dst[5] = src[5];
  dst[6] = src[6];
  dst[7] = src[7];
  return d;
}
#endif


#if (ACTUAL_SIZE_OF_C_LONG == 4 && NECESSARY_SIZE_OF_C_LONG == 4) || defined(GCCROBUG)
void tools_swaplong (long *n) {
  union {
    long d;
    unsigned char b[4];
  } p, q;
  p.d = *n;
  q.b[0] = p.b[3];
  q.b[1] = p.b[2];
  q.b[2] = p.b[1];
  q.b[3] = p.b[0];
  *n = q.d;
}
#elif (ACTUAL_SIZE_OF_C_LONG == 8 && NECESSARY_SIZE_OF_C_LONG == 8)  /* 0.31.7 patch for 64 bits architectures */
void tools_swaplong (long *n) {
  union {
    long d;
    unsigned char b[8];
  } p, q;
  p.d = *n;
  q.b[0] = p.b[7];
  q.b[1] = p.b[6];
  q.b[2] = p.b[5];
  q.b[3] = p.b[4];
  q.b[4] = p.b[3];
  q.b[5] = p.b[2];
  q.b[6] = p.b[1];
  q.b[7] = p.b[0];
  *n = q.d;
}
#else
void tools_swaplong (long *n) {
  int i;
  union {
    long d;
    unsigned char b[sizeof(long)];
  } p, q;
  p.d = *n;
  for (i=0; i < sizeof(long); i++) {
    q.b[i] = p.b[sizeof(long)-i-1];
  }
  *n = q.d;
}
#endif


void tools_swapint32_t (int32_t *n) {
  union {
    int32_t d;
    unsigned char b[4];
  } p, q;
  p.d = *n;
  q.b[0] = p.b[3];
  q.b[1] = p.b[2];
  q.b[2] = p.b[1];
  q.b[3] = p.b[0];
  *n = q.d;
}


/* complex sign */
double csgn (double a, double b) {
  if (a > 0 || (a == 0 && b > 0))
    return 1.0;
  else if (a < 0 || (a == 0 && b < 0))
    return -1.0;
  else
    return 0.0;
}


char *charreplace (char *s, char from, char to) {
  char *olds, *e;
  olds = s;
  while (NULL != (e = strchr(s, from))) {  /* traverses the entire string */
    s = e;
    *s = to;
  }
  s = s + strlen(s) - 1;  /* set pointer to end of string */
  /* delete trailing slash if present, only if str has more than one char, 0.31.4 */
  if (strlen(olds) > 1 && (*s == '/' || *s == '\\')) *s = '\0';
  s = olds;
  return s;
}


