/*
** $Id: biniolib.c,v 0.1 13.07.2008 $
** Library for reading and writing binary files
** See Copyright Notice in agena.h
*/

#include <stdio.h>      /* also for BUFSIZ */
#include <stdlib.h>
#include <string.h>
#include <unistd.h>     /* for ftruncate and close */
#include <fcntl.h>      /* for open */
#include <sys/stat.h>   /* S_I* constants */
#include <math.h>       /* for HUGE_VAL */
#include "agncfg.h"     /* for Endianness */

#define binio_c
#define LUA_LIB

#include "agena.h"
#include "agnxlib.h"
#include "agenalib.h"
#include "lobject.h"

#include "agncmpt.h"


static int binio_open (lua_State *L) {
  int hnd, readwrite;
  const char *file = agn_checkstring(L, 1);  /* 0.13.4 patch: does not accept numbers */
  /* no option given: open in readwrite mode, option given: open in readonly mode */
  readwrite = (lua_gettop(L) == 1);
  /* if the file already exists, check whether you have the proper file rights */
  if (access(file, F_OK) == 0) {  /* file already exists ? */
    if (access(file, (readwrite) ? R_OK|W_OK : R_OK) == -1)
      luaL_error(L, "Error in " LUA_QS ": missing permissions for " LUA_QS ".", "binio.open", file);
  }
  hnd = (readwrite) ? my_open(file) : my_roopen(file);
  if (hnd == -1) {  /* file does not yet exist ? */
    if (readwrite) {  /* a new file shall be created ? */
      hnd = my_create(file);
      if (hnd < 0)
        luaL_error(L, "Error in " LUA_QS ": " LUA_QS " could not be created.", "binio.open", file);
    } else
      luaL_error(L, "Error in " LUA_QS ": " LUA_QS " could not be opened.", "binio.open", file);
  }
  /* enter new open file to global binio.openfiles table */
  lua_getglobal(L, "binio");
  lua_getfield(L, -1, "openfiles");
  if (lua_istable(L, -1)) {  /* 0.20.2, avoid segmentation faults */
    lua_pushinteger(L, hnd);
    lua_pushstring(L, file);
    lua_rawset(L, -3);
  }
  agn_poptoptwo(L);  /* delete "binio" and "openfiles" */
  lua_pushinteger(L, hnd);
  return 1;
}


static int binio_close (lua_State *L) {
  int nargs, hnd, i, result;
  result = 1;
  nargs = lua_gettop(L);
  luaL_checkstack(L, nargs, "too many arguments");
  for (i=1; i <= nargs; i++) {
    hnd = luaL_checkinteger(L, i);
    if (hnd < 3) {  /* avoid standard streams to be closed */
      result = 0;
      continue;
    }
    if (!my_close(hnd) == 0)
      result = 0;
    else {
      /* delete file from global binio.openfiles table */
      lua_getglobal(L, "binio");
      lua_getfield(L, -1, "openfiles");
      if (lua_istable(L, -1)) {  /* 0.20.2, avoid segmentation faults */
        lua_pushinteger(L, hnd);
        lua_pushnil(L);
        lua_rawset(L, -3);
      }
      agn_poptoptwo(L);  /* delete "binio" and "openfiles" */
    }
  }
  if (result)
    lua_pushtrue(L);
  else
    luaL_error(L, "Error in " LUA_QS " with file #%d.", "binio.close", hnd);
  return 1;
}


static int binio_readchar (lua_State *L) {
  int32_t pos;
  int hnd, res;
  unsigned char data;
  hnd = luaL_checkinteger(L, 1);
  pos = luaL_optinteger(L, 2, 0);
  if (pos != 0 && lseek64(hnd, pos, SEEK_CUR) == -1) {
    lua_pushfail(L);
    return 1;
  }
  res = read(hnd, &data, sizeof(char));
  if (res > 0)
    lua_pushnumber(L, data);
  else if (res == 0)
    lua_pushnil(L);
  else
    luaL_error(L, "Error in " LUA_QS " with file #%d.", "binio.readchar", hnd);
  return 1;
}


/* 0.25.0, 18.07.2009; reads n bytes from a file and returns a sequence of numbers, i.e. the bytes read;
   patched 20.07.2009, 0.25.1 */

static int binio_readbytes (lua_State *L) {
  int32_t n, i;
  int hnd, res;
  hnd = luaL_checkinteger(L, 1);
  if (lua_gettop(L) == 1) {
    agn_getenv(L, "BufferSize");
    n = ( lua_type(L, -1) == LUA_TNUMBER ) ? agn_tonumber(L, -1) : BUFSIZ;
    agn_poptop(L);  /* pop Buffersize, 0.32.0 */
  } else
    n = agn_checknumber(L, 2);
  if (n < 1) n = BUFSIZ;
  unsigned char buffer[n];
  res = read(hnd, buffer, n);
  if (res > 0) {
    agn_createseq(L, res);
    for (i=0; i < res; i++)
      lua_seqsetinumber(L, -1, i+1, buffer[i]);
  }
  else if (res == 0)
    lua_pushnil(L);
  else
    luaL_error(L, "Error in " LUA_QS " with file #%d.", "binio.readbytes", hnd);
  return 1;
}


/* base_sync: flushes all unwritten content to the given file.
   Patched December 26, 2009, 0.30.0 */

static int binio_sync (lua_State *L) {
  int hnd, res;
  hnd = luaL_checkinteger(L, 1);
  #ifdef _WIN32
  res = _commit(hnd);
  #else
  res = fsync(hnd);
  #endif
  if (!res)
    lua_pushboolean(L, 1);
  else
    luaL_error(L, "Error in " LUA_QS " with file #%d.", "binio.sync", hnd);
  return 1;
}


static int binio_filepos (lua_State *L) {
  int hnd;
  off64_t fpos;
  hnd = luaL_checkinteger(L, 1);
  fpos = my_fpos(hnd);
  if (fpos == -1)
    luaL_error(L, "Error in " LUA_QS " with file #%d.", "binio.filepos", hnd);
  else
    lua_pushnumber(L, fpos);
  return 1;
}


static int binio_seek (lua_State *L) {
  int hnd;
  off64_t pos;
  hnd = luaL_checkinteger(L, 1);
  pos = luaL_checkinteger(L, 2);
  if (lseek64(hnd, pos, SEEK_CUR) == -1)
    luaL_error(L, "Error in " LUA_QS " with file #%d.", "binio.seek", hnd);
  else
    lua_pushtrue(L);
  return 1;
}


static int binio_length (lua_State *L) {
  int hnd;
  off64_t size, oldpos;
  hnd = luaL_checkinteger(L, 1);
  oldpos = lseek64(hnd, 0L, SEEK_CUR);
  size = lseek64(hnd, 0L, SEEK_END);
  lseek64(hnd, oldpos, SEEK_SET);  /* reset cursor to original file position */
  if (size == -1)
    luaL_error(L, "Error in " LUA_QS " with file #%d.", "binio.length", hnd);
  else
    lua_pushnumber(L, size);
  return 1;
}


static int binio_writechar (lua_State *L) {
  int hnd;
  size_t size;
  unsigned char data;
  hnd = luaL_checkinteger(L, 1);
  data = agn_checknumber(L, 2);
  size = sizeof(data);
  if (write(hnd, &data, size) != size)
    luaL_error(L, "Error in " LUA_QS " with writing file #%d.", "binio.writechar", hnd);
  else
    lua_pushtrue(L);
  return 1;
}


/* writes bytes stored in a sequence to a file, 0.25.0, 19.07.2009 */
static int binio_writebytes (lua_State *L) {
  int hnd;
  size_t size, i, nargs;
  hnd = luaL_checkinteger(L, 1);
  luaL_argcheck(L, lua_type(L, 2) == LUA_TSEQ, 2, "sequence expected");
  nargs = agn_seqsize(L, 2);
  if (nargs == 0)
    lua_pushfail(L);  /* better sure than sorry */
  else {
    unsigned char buffer[nargs];
    size = sizeof(buffer);
    for (i=0; i < nargs; i++)
      buffer[i] = (unsigned char)lua_seqgetinumber(L, 2, i+1);
    if (write(hnd, &buffer, size) != size)
      luaL_error(L, "Error in " LUA_QS " with writing file #%d.", "binio.writebytes", hnd);
    else
      lua_pushtrue(L);
  }
  return 1;
}


static int binio_writenumber (lua_State *L) {
  int hnd;
  size_t size;
  lua_Number data;
  uint64_t ulong;
  hnd = luaL_checkinteger(L, 1);
  data = agn_checknumber(L, 2);
  ulong = tools_swapdouble(data);
  size = sizeof(ulong);
  if (write(hnd, &ulong, size) != size)
    luaL_error(L, "Error in " LUA_QS " with writing file #%d.", "binio.writenumber", hnd);
  else
    lua_pushtrue(L);
  return 1;
}


static int binio_writelong (lua_State *L) {
  int hnd;
  size_t size;
  int32_t data;
  hnd = luaL_checkinteger(L, 1);
  data = agn_checknumber(L, 2);
  #if BYTE_ORDER != BIG_ENDIAN
  tools_swapint32_t(&data);
  #endif
  size = sizeof(data);
  if (write(hnd, &data, size) != size)
    luaL_error(L, "Error in " LUA_QS " with writing file #%d.", "binio.writelong", hnd);
  else
    lua_pushtrue(L);
  return 1;
}


static int binio_writestring (lua_State *L) {
  int hnd;
  size_t l;
  const char *value;
  hnd = luaL_checkinteger(L, 1);
  value = agn_checklstring(L, 2, &l);
  char data[l];
  strcpy(data, value);
  my_writel(hnd, l);
  if (write(hnd, &data, l) != l)
    luaL_error(L, "Error in " LUA_QS " with writing file #%d.", "binio.writestring", hnd);
  else
    lua_pushtrue(L);
  return 1;
}


static int binio_writeshortstring (lua_State *L) {  /* 0.32.0 */
  int hnd;
  size_t l;
  const char *value;
  hnd = luaL_checkinteger(L, 1);
  value = agn_checklstring(L, 2, &l);
  if (l > 255)
    luaL_error(L, "Error in " LUA_QS " with writing file #%d: string too long", "binio.writestring", hnd);
  char data[l];
  strcpy(data, value);
  my_writec(hnd, (unsigned char)l);
  if (write(hnd, &data, l) != l)
    luaL_error(L, "Error in " LUA_QS " with writing file #%d.", "binio.writestring", hnd);
  else
    lua_pushtrue(L);
  return 1;
}


static int binio_readstring (lua_State *L) {
  int hnd;
  ssize_t success;
  size_t size;
  hnd = luaL_checkinteger(L, 1);
  size = sec_readl(hnd, &success);  /* first read length of the string */
  if (success) {
    char data[size+1];
    if (read(hnd, &data, size) == size) {  /* now read string itself */
      data[size] = '\0';
      lua_pushlstring(L, data, size);
      return 1;
    }
  }
  luaL_error(L, "Error in " LUA_QS " with reading file #%d.", "binio.readstring", hnd);
  return 1;
}


static int binio_readshortstring (lua_State *L) {
  int hnd;
  size_t size;
  hnd = luaL_checkinteger(L, 1);
  if (sec_read(hnd, &size, sizeof(unsigned char))) {  /* first read length of the string */
    char data[size+1];
    if (read(hnd, &data, size) == size) {  /* now read string itself */
      data[size] = '\0';
      lua_pushlstring(L, data, size);
      return 1;
    }
  }
  luaL_error(L, "Error in " LUA_QS " with reading file #%d.", "binio.readstring", hnd);
  return 1;
}


static int binio_readnumber (lua_State *L) {
  int hnd;
  size_t success;
  uint64_t data;
  lua_Number num;
  hnd = luaL_checkinteger(L, 1);
  success = sec_read(hnd, &data, sizeof(uint64_t));
  if (success) {
    num = tools_unswapdouble(data);
    lua_pushnumber(L, num);
    return 1;
  }
  luaL_error(L, "Error in " LUA_QS " with reading file #%d.", "binio.readnumber", hnd);
  return 1;
}


static int binio_readlong (lua_State *L) {
  int hnd;
  size_t success;
  int32_t data;
  hnd = luaL_checkinteger(L, 1);
  success = sec_read(hnd, &data, sizeof(int32_t));
  if (success) {
    #if BYTE_ORDER != BIG_ENDIAN
    tools_swapint32_t(&data);
    #endif
    lua_pushnumber(L, data);
    return 1;
  }
  luaL_error(L, "Error in " LUA_QS " with reading file #%d.", "binio.readlong", hnd);
  return 1;
}


static int binio_toend (lua_State *L) {
  int hnd;
  off64_t size;
  hnd = luaL_checkinteger(L, 1);
  size = lseek64(hnd, 0L, SEEK_END);
  if (size == -1)
    luaL_error(L, "Error in " LUA_QS " with file #%d.", "binio.toend", hnd);
  else
    lua_pushnumber(L, size);
  return 1;
}


static int binio_rewind (lua_State *L) {
  int hnd;
  off64_t size;
  hnd = luaL_checkinteger(L, 1);
  size = lseek64(hnd, 0L, SEEK_SET);
  if (size == -1)
    luaL_error(L, "Error in " LUA_QS " with file #%d.", "binio.rewind", hnd);
  else
    lua_pushnumber(L, size);
  return 1;
}


static void binio_readobjecterror (lua_State *L, int hnd, int message) {  /* May 03, 2010 */
  /* close the file and delete file name from global binio.openfiles table */
  if (my_close(hnd) == 0) {
    lua_getglobal(L, "binio");
    lua_getfield(L, -1, "openfiles");
    if (lua_istable(L, -1)) {  /* 0.20.2, avoid segmentation faults */
      lua_pushinteger(L, hnd);
      lua_pushnil(L);
      lua_rawset(L, -3);
    }
    agn_poptoptwo(L);  /* delete "binio" and "openfiles" */
  }
  if (message)
    luaL_error(L, "Error in " LUA_QS " with reading file #%d: I/O error.", "binio.readobject", hnd);
  else
    luaL_error(L, "Error in " LUA_QS ": wrong workspace file version.", "binio.readobject");
}


static void binio_readobjectutypename (lua_State *L, int hnd, int idx) {
  size_t size;
  ssize_t success;
  size = sec_readl(hnd, &success);  /* first read length of the string */
  if (success) {
    char data[size+1];
    if (read(hnd, &data, size) == size) {  /* now read string itself */
      data[size] = '\0';
      lua_pushlstring(L, data, size);
      agn_setutype(L, idx-1, -1);  /* set user-defined type to set */
      agn_poptop(L);            /* delete utype string */
      return;
    }  /* else issue an error */
  }
  binio_readobjecterror(L, hnd, 1);
}


static void binio_readobjectaux (lua_State *L, int hnd) {  /* May 03, 2010 */
  unsigned char data;
  int res;
  res = read(hnd, &data, sizeof(char));
  if (res > 0) {
    switch (data - (data & 32)) {  /* get data type ID */
      case 0: {  /* read char */
        unsigned char num;
        if (sec_read(hnd, &num, sizeof(unsigned char)))
          lua_pushnumber(L, num);
        else
          binio_readobjecterror(L, hnd, 1);
        break;
      }
      case 1: {  /* read long */
        int32_t data;
        if (sec_read(hnd, &data, sizeof(int32_t))) {
          #if BYTE_ORDER != BIG_ENDIAN
          tools_swapint32_t(&data);
          #endif
          lua_pushnumber(L, data);
        } else
          binio_readobjecterror(L, hnd, 1);
        break;
      }
      case 2: {  /* read number */
        lua_Number n;
        uint64_t num;
        if (sec_read(hnd, &num, sizeof(uint64_t))) {
          n = tools_unswapdouble(num);
          lua_pushnumber(L, n);
        } else
          binio_readobjecterror(L, hnd, 1);
        break;
      }
      case 3: {  /* read string */
        size_t size;
        ssize_t success;
        size = sec_readl(hnd, &success);  /* first read length of the string */
        if (success) {
          char data[size+1];
          if (read(hnd, &data, size) == size) {  /* now read string itself */
            data[size] = '\0';
            lua_pushlstring(L, data, size);
            return;
          }
        }
        binio_readobjecterror(L, hnd, 1);
        break;
      }
      case 4: {  /* read short string < 256 chars */
        unsigned char size;
        if (sec_read(hnd, &size, sizeof(unsigned char))) {
          char data[size+1];
          if (read(hnd, &data, size) == size) {  /* now read string itself */
            data[size] = '\0';
            lua_pushlstring(L, data, size);
            return;
          }
        }
        binio_readobjecterror(L, hnd, 1);
        break;
      }
      case 5: {  /* read table */
        lua_Number i, size;
        uint64_t num;
        if (sec_read(hnd, &num, sizeof(uint64_t))) {
          size = tools_unswapdouble(num);
          lua_newtable(L);
          if ((data & 32) == 32)  /* has a user-defined type been set ? */
            binio_readobjectutypename(L, hnd, -1);
          for (i=1; i <= size; i++) {
            binio_readobjectaux(L, hnd);
            binio_readobjectaux(L, hnd);
            lua_rawset(L, -3);
          }
        } else
          binio_readobjecterror(L, hnd, 1);
        break;
      }
      case 6: {  /* read set */
        lua_Number i, size;
        uint64_t num;
        if (sec_read(hnd, &num, sizeof(uint64_t))) {
          size = tools_unswapdouble(num);
          agn_createset(L, size);
          for (i=1; i <= size; i++) {
            binio_readobjectaux(L, hnd);
            lua_srawset(L, -2);
          }
        } else
          binio_readobjecterror(L, hnd, 1);
        break;
      }
      case 7: {  /* read sequence */
        lua_Number i, size;
        uint64_t num;
        if (sec_read(hnd, &num, sizeof(uint64_t))) {
          size = tools_unswapdouble(num);
          agn_createseq(L, size);
          if ((data & 32) == 32)  /* has a user-defined type been set ? */
            binio_readobjectutypename(L, hnd, -1);
          /* read contents of sequence */
          for (i=1; i <= size; i++) {
            binio_readobjectaux(L, hnd);
            lua_seqinsert(L, -2);
          }
        } else
          binio_readobjecterror(L, hnd, 1);
        break;
      }
      case 8: {  /* read pair */
        int typeset = ((data & 32) == 32);
        if (typeset) {  /* read user defined type */
          size_t size;
          ssize_t success;
          size = sec_readl(hnd, &success);  /* first read length of the string */
          if (success) {
            char data[size+1];
            if (read(hnd, &data, size) == size) {  /* now read string itself */
              data[size] = '\0';
              lua_pushlstring(L, data, size);
            } else
              binio_readobjecterror(L, hnd, 1);
          } else
            binio_readobjecterror(L, hnd, 1);
        }
        /* read pair */
        binio_readobjectaux(L, hnd);
        binio_readobjectaux(L, hnd);
        agn_createpair(L, -2, -1);  /* assign it */
        lua_remove(L, -2);  /* remove left .. */
        lua_remove(L, -2);  /* .. and right hand-side */
        if (typeset) {  /* now assign user-defined type */
          agn_setutype(L, -1, -2);
          lua_remove(L, -2);  /* and remove utype string */
        }
        break;
      }
      case 9: {  /* read complex */
        lua_Number r, i;
        binio_readobjectaux(L, hnd);  /* read real part (long or double) */
        binio_readobjectaux(L, hnd);  /* read imaginary part (long or double) */
        r = lua_tonumber(L, -2);
        i = lua_tonumber(L, -1);
        agn_poptoptwo(L);
#ifndef PROPCMPLX
        agn_createcomplex(L, r+I*i);
        #else
        agn_createcomplex(L, r, i);
        #endif
        break;
      }
      case 10: {  /* read boolean */
        unsigned char type;
        if (read(hnd, &type, sizeof(char)) < 1)
          binio_readobjecterror(L, hnd, 1);
        else {
          if (type == 2)
            lua_pushfail(L);
          else
            lua_pushboolean(L, type);
        }
        break;
      }
      case 11: {  /* read null */
        lua_pushnil(L);
        break;
      }
      default:
        luaL_error(L, "Error in " LUA_QS " with reading file #%d: unknown type.", "binio.readobject", hnd);
    }
  }
  else if (res == 0)
    lua_pushnil(L);
  else
    binio_readobjecterror(L, hnd, 1);
}


#define FILESPEC "AGENAWORKSPACE"
#define FILESPECLEN 14
#define FILESPECVER 1
#define FILESPECSUB 0

static int binio_readobject (lua_State *L) {
  int hnd;
  hnd = luaL_checkinteger(L, 1);
  if (lseek64(hnd, 0L, SEEK_CUR) == -1)
    binio_readobjecterror(L, hnd, 1);
  else {
    char data[FILESPECLEN+1];
    if (read(hnd, &data, FILESPECLEN) == FILESPECLEN) {
      unsigned char version, subversion;
      data[FILESPECLEN] = '\0';
      /* read file header */
      if (strcmp(data, FILESPEC) != 0)
        binio_readobjecterror(L, hnd, 0);
      /* read file version and subversion */
      if (sec_read(hnd, &version, sizeof(unsigned char)) && (version != FILESPECVER))
        binio_readobjecterror(L, hnd, 0);
      if (sec_read(hnd, &subversion, sizeof(unsigned char)) && (subversion < FILESPECSUB))
        binio_readobjecterror(L, hnd, 0);
    } else
      binio_readobjecterror(L, hnd, 1);
    binio_readobjectaux(L, hnd);
  }
  return 1;
}


static int binio_lock (lua_State *L) {  /* 0.32.5 */
  int hnd;
  size_t nargs;
  off64_t start, size;
  hnd = luaL_checkinteger(L, 1);
  nargs = lua_gettop(L);
  luaL_checkstack(L, nargs, "too many arguments");
  if (nargs == 1) {  /* lock entire file (in Windows lock 2^63 bytes only) */
    start = 0;
    size = 0;
  } else {
    /* lock from current file position */
    start = my_fpos(hnd);
    size = luaL_checknumber(L, 2);
    if (size < 0) luaL_error(L, "Error in " LUA_QS ": must lock at least one byte.", "binio.lock");
  }
  lua_pushboolean(L, my_lock(hnd, start, size) == 0);
  return 1;
}


static int binio_unlock (lua_State *L) {  /* 0.32.5 */
  int hnd;
  size_t nargs;
  off64_t start, size;
  hnd = luaL_checkinteger(L, 1);
  nargs = lua_gettop(L);
  luaL_checkstack(L, nargs, "too many arguments");
  if (nargs == 1) {  /* lock entire file */
    start = 0;
    size = 0;
  } else {
    /* lock from current file position */
    start = my_fpos(hnd);
    size = luaL_checknumber(L, 2);
    if (size < 0) luaL_error(L, "Error in " LUA_QS ": must unlock at least one byte.", "binio.unlock");
  }
  lua_pushboolean(L, my_unlock(hnd, start, size) == 0);
  return 1;
}


static const luaL_Reg binio[] = {
  {"close", binio_close},                  /* July 13, 2008 */
  {"filepos", binio_filepos},              /* July 13, 2008 */
  {"length", binio_length},                /* July 13, 2008 */
  {"lock", binio_lock},                    /* June 12, 2010 */
  {"open", binio_open},                    /* July 13, 2008 */
  {"readbytes", binio_readbytes},          /* July 18, 2009 */
  {"readchar", binio_readchar},            /* July 13, 2008 */
  {"readlong", binio_readlong},            /* October 26, 2008 */
  {"readnumber", binio_readnumber},        /* October 25, 2008 */
  {"readstring", binio_readstring},        /* July 20, 2008 */
  {"readshortstring", binio_readshortstring},  /* May 08, 2010 */
  {"rewind", binio_rewind},                /* July 20, 2008 */
  {"seek", binio_seek},                    /* July 13, 2008 */
  {"sync", binio_sync},                    /* July 13, 2008 */
  {"toend", binio_toend},                  /* July 20, 2008 */
  {"unlock", binio_unlock},                /* June 12, 2010 */
  {"writebytes", binio_writebytes},        /* July 19, 2009 */
  {"writechar", binio_writechar},          /* July 19, 2008 */
  {"writelong", binio_writelong},          /* October 26, 2008 */
  {"writenumber", binio_writenumber},      /* October 25, 2008 */
  {"writestring", binio_writestring},      /* July 20, 2008 */
  {"writeshortstring", binio_writeshortstring},  /* May 08, 2010 */
  {"readobject", binio_readobject},      /* May 02, 2010 */
  {NULL, NULL}
};


static void createmetatable (lua_State *L) {
  lua_createtable(L, 0, 1);  /* create metatable for strings */
  lua_pushliteral(L, "");  /* dummy string */
  lua_pushvalue(L, -2);
  lua_setmetatable(L, -2);  /* set string metatable */
  agn_poptop(L);  /* pop dummy string */
  lua_pushvalue(L, -2);  /* string library... */
  lua_setfield(L, -2, "__index");  /* ...is the __index metamethod */
  agn_poptop(L);  /* pop metatable */
}


/*
** Open binio library
*/
LUALIB_API int luaopen_binio (lua_State *L) {
  luaL_register(L, AGENA_BINIOLIBNAME, binio);
  lua_newtable(L);
  lua_setfield(L, -2, "openfiles");  /* table for information on all open files */
  createmetatable(L);
  return 1;
}

