/*
** $Id: loslib.c,v 1.19 2006/04/26 18:19:49 roberto Exp $
** Standard Operating System library
** See Copyright Notice in agena.h
*/


#include <errno.h>
#include <locale.h>
#include <stdlib.h>
#include <string.h>

/* #include <time.h> */
#include "agnt64.h"

#include <sys/stat.h> /* for mkdir, stat */

/* for os.fstat to determine correct sizes of files > 4 GB */
#ifdef _WIN32
  #undef stat
  #define stat  _stati64
  #undef fstat
  #define fstat _fstati64
  #undef wstat
  #define wstat _wstati64
#elif defined (__SVR4) && defined (__sun)
  #undef stat
  #define stat stat64
#elif defined(__unix__) && !defined(LUA_DOS)
  #undef _FILE_OFFSET_BITS
  #define _FILE_OFFSET_BITS	64
#endif


#if _WIN32
#include <windows.h>  /* for os_memstate */
#include <io.h>       /* for access */
#include <ctype.h>    /* for toupper */
#endif

/* This OS2 header must be placed before including unistd, utsname, ncurses,
   otherwise GCC will not compile successfully. 0.13.3 */
#ifdef __OS2__
#define INCL_DOS
#define INCL_DOSERRORS
#define INCL_NOPM
#define INCL_DOSPROCESS  /* for DosBeep */
#define INCL_DOSFILEMGR
#include <os2.h>     /* for memory status queries */
#endif

#if defined(__unix__) || defined(__OS2__) || defined(__APPLE__) || defined(__HAIKU__)
#include <unistd.h>   /* for access and sleep */
#include <sys/utsname.h>  /* for uname */
#endif

#ifdef __APPLE__
#include <mach/mach.h>
#include <sys/sysctl.h>
#include "agncmpt.h"
#endif

#include <dirent.h>  /* used by io_ls function */
#include <errno.h>   /* used by io_ls function */

#include "agnhlps.h"

#define loslib_c
#define LUA_LIB

#include "agena.h"

#include "agnxlib.h"
#include "agenalib.h"


static int os_pushresult (lua_State *L, int i, const char *filename) {
  int en = errno;  /* calls to Lua API may change this value */
  if (i == -1) {  /* function failed ? */
    lua_pushfail(L);
    lua_pushfstring(L, "%s: %s", filename, strerror(en));
    lua_pushinteger(L, en);
    return 3;
  }
  else {
    lua_pushtrue(L);
    return 1;
  }
}


static void setboolean (lua_State *L, const char *key, unsigned int value) {
  lua_pushstring(L, key);
  lua_pushboolean(L, value);
  lua_settable(L, -3);
}


static void setstring (lua_State *L, const char *key, const char *value) {
  lua_pushstring(L, key);
  lua_pushstring(L, value);
  lua_rawset(L, -3);
}


static void setnumber (lua_State *L, const char *key, lua_Number value) {
  lua_pushstring(L, key);
  lua_pushnumber(L, value);
  lua_rawset(L, -3);
}


static int os_execute (lua_State *L) {
  lua_pushinteger(L, system(luaL_optstring(L, 1, NULL)));
  return 1;
}


static int os_remove (lua_State *L) {
  const char *filename = luaL_checkstring(L, 1);
  if (access(filename, 00|04) != -1)  /* 0.26.0 patch */
    return os_pushresult(L, remove(filename) == 0, filename);
  else {
    lua_pushfail(L);
    return 1;
  }
}


static int os_move (lua_State *L) {
  const char *fromname = luaL_checkstring(L, 1);
  const char *toname = luaL_checkstring(L, 2);
  if (access(fromname, 00|04) != -1)  /* 0.26.0 patch */
    return os_pushresult(L, rename(fromname, toname) == 0, fromname);
  else {
    lua_pushfail(L);
    return 1;
  }
}


static int os_tmpname (lua_State *L) {
  char buff[LUA_TMPNAMBUFSIZE];
  int err;
  lua_tmpnam(buff, err);
  if (err)
    return luaL_error(L, "unable to generate a unique filename.");
  lua_pushstring(L, buff);
  return 1;
}


static int os_getenv (lua_State *L) {
  lua_pushstring(L, getenv(luaL_checkstring(L, 1)));  /* if NULL push nil */
  return 1;
}


/*
** {======================================================
** Time/Date operations
** { year=%Y, month=%m, day=%d, hour=%H, min=%M, sec=%S,
**   wday=%w+1, yday=%j, isdst=? }
** =======================================================
*/

static void setfield (lua_State *L, const char *key, int value) {
  lua_pushinteger(L, value);
  lua_setfield(L, -2, key);
}

static void setboolfield (lua_State *L, const char *key, int value) {
  if (value < 0)  /* undefined? */
    return;  /* does not set field */
  lua_pushboolean(L, value);
  lua_setfield(L, -2, key);
}

static int getboolfield (lua_State *L, const char *key) {
  int res;
  lua_getfield(L, -1, key);
  res = lua_isnil(L, -1) ? -1 : lua_toboolean(L, -1);
  agn_poptop(L);
  return res;
}


static int getfield (lua_State *L, const char *key, int d) {
  int res;
  lua_getfield(L, -1, key);
  if (lua_isnumber(L, -1))
    res = (int)lua_tointeger(L, -1);
  else {
    if (d < 0)
      return luaL_error(L, "field " LUA_QS " missing in date table.", key);
    res = d;
  }
  agn_poptop(L);
  return res;
}


static int os_date (lua_State *L) {
  const char *s = luaL_optstring(L, 1, "%c");
  Time64_T t = luaL_opt(L, (Time64_T)luaL_checknumber, 2, time(NULL));
  struct TM *stm;
  if (*s == '!') {  /* UTC? */
    stm = gmtime64(&t);
    s++;  /* skip `!' */
  }
  else
    stm = localtime64(&t);
  if (stm == NULL)  /* invalid date? */
    lua_pushnil(L);
  else if (strcmp(s, "*t") == 0) {
    lua_createtable(L, 0, 9);  /* 9 = number of fields */
    setfield(L, "sec", stm->tm_sec);
    setfield(L, "min", stm->tm_min);
    setfield(L, "hour", stm->tm_hour);
    setfield(L, "day", stm->tm_mday);
    setfield(L, "month", stm->tm_mon+1);
    setfield(L, "year", stm->tm_year+1900);
    setfield(L, "wday", stm->tm_wday+1);
    setfield(L, "yday", stm->tm_yday+1);
    setboolfield(L, "isdst", stm->tm_isdst);
  }
  else {  /* Lua 5.1.2 patch */
    char cc[3];
    luaL_Buffer b;
    cc[0] = '%'; cc[2] = '\0';
    luaL_buffinit(L, &b);
    for (; *s; s++) {
      if (*s != '%' || *(s + 1) == '\0')  /* no conversion specifier? */
        luaL_addchar(&b, *s);
      else {
        size_t reslen;
        char buff[256];  /* should be big enough for any conversion result */
        cc[1] = *(++s);
        reslen = strftime(buff, sizeof(buff), cc, stm);
        luaL_addlstring(&b, buff, reslen);
      }
    }
    luaL_pushresult(&b);
  }
  return 1;
}


static int os_time (lua_State *L) {
  time_t t;
  if (lua_isnoneornil(L, 1))  /* called without args? */
    t = time(NULL);  /* get current time */
  else {
    struct tm ts;
    luaL_checktype(L, 1, LUA_TTABLE);
    lua_settop(L, 1);  /* make sure table is at the top */
    ts.tm_sec = getfield(L, "sec", 0);
    ts.tm_min = getfield(L, "min", 0);
    ts.tm_hour = getfield(L, "hour", 12);
    ts.tm_mday = getfield(L, "day", -1);
    ts.tm_mon = getfield(L, "month", -1) - 1;
    ts.tm_year = getfield(L, "year", -1) - 1900;
    ts.tm_isdst = getboolfield(L, "isdst");
    t = mktime(&ts);
  }
  if (t == (time_t)(-1))
    lua_pushnil(L);
  else
    lua_pushnumber(L, (lua_Number)t);
  return 1;
}


static int os_difftime (lua_State *L) {
  lua_pushnumber(L, difftime((time_t)(luaL_checknumber(L, 1)),
                             (time_t)(luaL_optnumber(L, 2, 0))));
  return 1;
}

/* }====================================================== */


static int os_setlocale (lua_State *L) {
  static const int cat[] = {LC_ALL, LC_COLLATE, LC_CTYPE, LC_MONETARY,
                      LC_NUMERIC, LC_TIME};
  static const char *const catnames[] = {"all", "collate", "ctype", "monetary",
     "numeric", "time", NULL};
  const char *l = luaL_optstring(L, 1, NULL);
  int op = luaL_checkoption(L, 2, "all", catnames);
  lua_pushstring(L, setlocale(cat[op], l));
  return 1;
}


static int os_exit (lua_State *L) {
  exit(luaL_optint(L, 1, EXIT_SUCCESS));
  return 0;  /* to avoid warnings */
}


/* determine information on available memory on Windows platforms, 0.5.3, September 15, 2007;
   extended in 0.12.2 on September 29, 2008 to work in UNIX; extended to fully support OS/2 in 0.13.3;
   extended to support Mac OS X in 0.31.3 */

static int os_memstate (lua_State *L) {
#if defined(_WIN32) || (defined(__unix__) && !defined(LUA_DOS)) || defined(__OS2__) || defined(__APPLE__) || defined(__HAIKU__)
  long AvailPhys, TotalPhys;
  lua_Number div;
#if defined(_WIN32)
  MEMORYSTATUS memstat;
  long AvailVirtual, TotalVirtual, MemoryLoad;
  int Length;
#elif __OS2__
  APIRET result;
  ULONG memInfo[3];
#elif defined(__APPLE__)
  vm_size_t pagesize;
  vm_statistics_data_t vminfo;
  mach_port_t system;
  unsigned int c;
  int sysmask[2];
  size_t length;
  uint64_t memsize;
  lua_Number f;
#endif
  const char *option = luaL_optstring(L, 1, "bytes");
  if (strcmp(option, "bytes") == 0)
    div = 1;
  else if (strcmp(option, "kbytes") == 0)
    div = 1024;
  else if (strcmp(option, "mbytes") == 0)
    div = 1024*1024;
  else if (strcmp(option, "gbytes") == 0)
    div = 1024*1024*1024;
  else
    return luaL_error(L, "unknown option %s.", option);
#if defined(_WIN32)
  memstat.dwLength = sizeof(MEMORYSTATUS);
  GlobalMemoryStatus(&memstat);
  AvailPhys = (long)memstat.dwAvailPhys;
  TotalPhys = (long)memstat.dwTotalPhys;
  AvailVirtual = (long)memstat.dwAvailVirtual;
  TotalVirtual = (long)memstat.dwTotalVirtual;
  MemoryLoad = (long)memstat.dwMemoryLoad;
  Length = (int)memstat.dwLength;
  lua_newtable(L);
  setnumber(L, "freephysical", (lua_Number)(AvailPhys)/div);
  setnumber(L, "totalphysical", (lua_Number)(TotalPhys)/div);
  setnumber(L, "freevirtual", (lua_Number)(AvailVirtual)/div);
  setnumber(L, "totalvirtual", (lua_Number)(TotalVirtual)/div);  /* 0.26.0 patch */
#elif (defined(__unix__) && !defined(LUA_DOS)) || defined(__HAIKU__) /* for DJGPP */
  AvailPhys = sysconf(_SC_AVPHYS_PAGES) * sysconf(_SC_PAGESIZE);
  TotalPhys = sysconf(_SC_PHYS_PAGES) * sysconf(_SC_PAGESIZE);
  lua_newtable(L);
  setnumber(L, "freephysical", (lua_Number)(AvailPhys)/div);
  setnumber(L, "totalphysical", (lua_Number)(TotalPhys)/div);
#elif __OS2__
  result = DosQuerySysInfo(QSV_TOTPHYSMEM, QSV_TOTAVAILMEM,
      &memInfo[0], sizeof(memInfo));
  AvailPhys = 0;
  TotalPhys = 0;
    /* sysconf(_SC_AVPHYS_PAGES) * sysconf(_SC_PAGESIZE); does not work for
      the _SC* names are not defined in GCC unistd.h */
  if (result != 0) {
    lua_pushfail(L);
    return 1;
  }
  lua_newtable(L);
  setnumber(L, "totalphysical", (lua_Number)memInfo[0]/div);
  setnumber(L, "freevirtual", (lua_Number)memInfo[2]/div);
  setnumber(L, "resident", (lua_Number)memInfo[1]/div);
#elif defined(__APPLE__)
  sysmask[0] = CTL_HW;
  sysmask[1] = HW_MEMSIZE;
  length = sizeof(memsize);
  memsize = 0L;
  lua_newtable(L);
  c = HOST_VM_INFO_COUNT;
  system = mach_host_self();
  AvailPhys = AGN_NAN;  /* to prevent compiler warnings */
  TotalPhys = AGN_NAN;  /* to prevent compiler warnings */
  if (host_page_size(system, &pagesize) != KERN_SUCCESS) pagesize = 4096;
  f = pagesize/div;
  if (host_statistics(system, HOST_VM_INFO, (host_info_t)(&vminfo), &c) != KERN_SUCCESS) {
    setnumber(L, "freephysical", AvailPhys);
    setnumber(L, "active", AGN_NAN);
    setnumber(L, "inactive", AGN_NAN);
    setnumber(L, "speculative", AGN_NAN);
    setnumber(L, "wireddown", AGN_NAN);
    setnumber(L, "reactivated", AGN_NAN);
  }
  else {
    setnumber(L, "freephysical", ((vminfo.free_count - vminfo.speculative_count)) * f);
    setnumber(L, "active", vminfo.active_count * f);
    setnumber(L, "inactive", vminfo.inactive_count * f);
    setnumber(L, "speculative", vminfo.speculative_count * f);
    setnumber(L, "wireddown", vminfo.wire_count * f);
    setnumber(L, "reactivated", vminfo.reactivations * f);
  }
  if (sysctl(sysmask, 2, &memsize, &length, NULL, 0 ) == -1)
    setnumber(L, "totalphysical", TotalPhys);
  else
    setnumber(L, "totalphysical", memsize/div);
#endif
#else
  lua_pushfail(L);
#endif
  return 1;
}

/* get the current Windows login name, 0.5.3, September 15, 2007;
   extended 0.12.2, October 12, 2008; changed to support OS/2 0.13.3 */

static int os_login (lua_State *L) {
#if defined(_WIN32)
  char *LoginName;
  char buffer[200];
  DWORD len = 200;
  if (GetUserName(buffer, &len)) {
    LoginName = (char *)malloc(len+1);
    if (!LoginName)
      return luaL_error(L, "could not determine login name.\n");
    strncpy(LoginName, buffer, len);
    LoginName[len] = '\0';
  }
  else {
    LoginName = "<Unknown>";
  }
  lua_pushstring(L, LoginName);
#elif defined(__unix__) || defined(__OS2__) || defined(__APPLE__) || defined(__HAIKU__)
  char *LoginName;
  LoginName = getlogin();
  if (LoginName == NULL)
    lua_pushfail(L);
  else
    lua_pushstring(L, LoginName);
#else
  lua_pushfail(L);
#endif
  return 1;
}


/* get the Windows computer name, 0.5.3, September 15, 2007;
   improved in 0.13.3 */

static int os_computername (lua_State *L) {
#if defined(_WIN32)
  char *NTComputerName;
  char buffer[256];
  DWORD len = 256;
  if (GetComputerName(buffer, &len)) {
    NTComputerName = (char *)malloc(len+1);
    if (!NTComputerName)
      return luaL_error(L, "could not determine login name.\n");
    strncpy(NTComputerName, buffer, len);
    NTComputerName[len] = '\0';
  }
  else {
    NTComputerName = "<Unknown>";
  }
  lua_pushstring(L, NTComputerName);
#elif defined(__unix__) || defined(__APPLE__) || defined(__HAIKU__)
  /* does not work with OS/2: needs a file `TCPIP32` (as does bash). If not
     present, Agena does not start. */
  char buffer[256];
  size_t len = 256;
  if (gethostname(buffer, len) == 0)
    lua_pushstring(L, buffer);
  else
    lua_pushfail(L);
#else
  lua_pushfail(L);
#endif
  return 1;
}


/* return amount of free memory; extended 0.12.2, 28.09.2008; OS/2 addon
   August 07, 2009, OS X addon 0.31.3 March 28, 2010 */
static int os_freemem (lua_State *L) {
#if defined(_WIN32) || (defined(__unix__) && !defined(LUA_DOS)) || defined(__OS2__) || defined(__APPLE__) || defined(__HAIKU__)
  lua_Number div;
  const char *option = luaL_optstring(L, 1, "bytes");
#if defined(_WIN32)
  MEMORYSTATUS memstat;
#elif __OS2__
  APIRET result;
  ULONG memInfo[3];
#elif defined(__APPLE__)
  vm_size_t pagesize;
  vm_statistics_data_t vminfo;
  mach_port_t system;
  unsigned int c;
#endif
  if (strcmp(option, "bytes") == 0)
    div = 1;
  else if (strcmp(option, "kbytes") == 0)
    div = 1024;
  else if (strcmp(option, "mbytes") == 0)
    div = 1024*1024;
  else if (strcmp(option, "gbytes") == 0)
    div = 1024*1024*1024;
  else
    return luaL_error(L, "unknown option %s.", option);
#if defined(_WIN32)
  memstat.dwLength = sizeof(MEMORYSTATUS);
  GlobalMemoryStatus(&memstat);
  lua_pushnumber(L, (lua_Number)memstat.dwAvailPhys/div);
#elif defined(__unix__) && !defined(LUA_DOS) || defined(__HAIKU__)  /* excludes DJGPP */
  lua_pushnumber(L, (lua_Number)(sysconf(_SC_AVPHYS_PAGES) * sysconf(_SC_PAGESIZE))/div);
#elif defined (__OS2__)
  result = DosQuerySysInfo(QSV_TOTPHYSMEM, QSV_TOTAVAILMEM,
      &memInfo[0], sizeof(memInfo));
  if (result != 0)  /* return free virtual memory */
    lua_pushfail(L);
  else
    lua_pushnumber(L, (lua_Number)(memInfo[2])/div);
#elif defined(__APPLE__)
  c = HOST_VM_INFO_COUNT;
  system = mach_host_self();
  if (host_page_size(mach_host_self(), &pagesize) != KERN_SUCCESS) pagesize = 4096;
  if (host_statistics(system, HOST_VM_INFO, (host_info_t)(&vminfo), &c) != KERN_SUCCESS)
    lua_pushfail(L);
  else
    lua_pushnumber(L, (uint64_t)(vminfo.free_count - vminfo.speculative_count) * pagesize/div);
#endif
#else
  lua_pushfail(L);
#endif
  return 1;
}


static int os_listcore (lua_State *L) {
  DIR *dir;
  struct dirent *entry;
  int i, isdir, isfile, islink, nops, push;
  const char *path, *option, *fn;
#ifndef __WIN32
  struct stat o;
  mode_t mode;
#else
  DWORD mode;
#endif
  isdir = isfile = islink = 0;
  nops = lua_gettop(L);
  path = luaL_optstring(L, 1, tools_cwd());
  path = (const char *)charreplace((char *)path, '\\', '/');
  for (i=2; i <= nops; i++) {
    option = luaL_checkstring(L, i);
    if (strcmp(option, "files") == 0) isfile = 1;
    else if (strcmp(option, "dirs") == 0) isdir = 1;
    else if (strcmp(option, "links") == 0) islink = 1;
  }
  dir = opendir(path);
  if (dir == NULL) {
    lua_pushnil(L);
    lua_pushstring(L, strerror(errno));
    return 2;
  }
  lua_newtable(L);
  i = 1;
  while ((entry = readdir(dir)) != NULL) {
    push = nops <= 1;
    fn = entry->d_name;
    if (strcmp(fn, ".") == 0 || strcmp(fn, "..") == 0) continue;
    if (nops > 1) {
      int isslash;
      isslash = (strcmp(path, "/") == 0);
      if (!isslash)  /* 0.31.4, do not create `//...` */
        lua_pushstring(L, path);
      lua_pushstring(L, "/");
      lua_pushstring(L, fn);
      lua_concat(L, 3-isslash);
#ifdef __WIN32
      mode = GetFileAttributes(agn_tostring(L, -1));
      if (mode != INVALID_FILE_ATTRIBUTES)
        push = ((mode & FILE_ATTRIBUTE_DIRECTORY) && isdir) ||
               (!(mode & FILE_ATTRIBUTE_DIRECTORY) && isfile) ||
               ((mode & FILE_ATTRIBUTE_REPARSE_POINT) && islink);
      agn_poptop(L);
    }
#else
      if (stat(agn_tostring(L, -1), &o) == 0) {
        mode = o.st_mode;
        push = (
               (S_ISDIR(mode) && isdir)
               || (S_ISREG(mode) && isfile)
#if (defined(__unix__) && !defined(LUA_DOS)) || defined(__APPLE__) || defined(__HAIKU__)
               || (S_ISLNK(mode) && islink)
#endif
             );
      }
      agn_poptop(L);
    }
#endif
    if (push) {
      lua_pushnumber(L, i++);
      lua_pushstring(L, fn);
      lua_settable(L, -3);
    }
  }
  closedir(dir);
  return 1;
}


static int os_fstat (lua_State *L) {  /* 0.12.2, October 11, 2008; modified 0.26.0, August 06, 2009; patched 0.31.3 */
  struct stat entry;
  const char *fn;
  mode_t mode;
  struct tm *stm;
  int i;
  time_t seconds;
  char bits[17] = "----------:-----\0";
#ifdef _WIN32
  DWORD winmodet;
#elif defined(__OS2__)
  static FILESTATUS3 fstat, *fs;
  APIRET rc;
#endif
  i = 0;
  fn = luaL_checkstring(L, 1);
  if (stat(fn, &entry) == 0) {
    lua_newtable(L);
    mode = entry.st_mode;
    lua_pushstring(L, "mode");
    if (S_ISDIR(mode)) {
      lua_pushstring(L, "DIR");
      bits[0] = 'd';
#if defined(_WIN32) || defined(__OS2__) || defined (LUA_DOS)
      bits[11] = 'd';
#endif
    }
#if (defined(__unix__) && !defined(LUA_DOS)) || defined(__APPLE__) || defined(__HAIKU__)
    else if (S_ISLNK(mode)) {
      lua_pushstring(L, "LINK");
      bits[0] = 'l';
    }
#endif
    else if (S_ISREG(mode))
      lua_pushstring(L, "FILE");
    else if (S_ISCHR(mode))
      lua_pushstring(L, "CHARSPECFILE");
    else if (S_ISBLK(mode))
      lua_pushstring(L, "BLOCKSPECFILE");
    else
      lua_pushstring(L, "OTHER");
    lua_settable(L, -3);
    lua_pushstring(L, "length");
    lua_pushnumber(L, entry.st_size);
    lua_settable(L, -3);
    seconds = entry.st_mtime;
    lua_pushstring(L, "date");
    lua_createtable(L, 6, 0);  /* 6 = number of entries */
    stm = localtime(&seconds);
    if (stm != NULL) {  /* 0.31.3 patch */
      lua_rawsetinumber(L, -1, 1, stm->tm_year+1900);
      lua_rawsetinumber(L, -1, 2, stm->tm_mon+1);
      lua_rawsetinumber(L, -1, 3, stm->tm_mday);
      lua_rawsetinumber(L, -1, 4, stm->tm_hour);
      lua_rawsetinumber(L, -1, 5, stm->tm_min);
      lua_rawsetinumber(L, -1, 6, stm->tm_sec);
      lua_settable(L, -3);
    }
    /* 0.25.1 file attributes additions */
    /* determine owner rights */
    lua_pushstring(L, "user");
    lua_createtable(L, 0, 3);
    setboolean(L, "read", mode & S_IRUSR);
    setboolean(L, "write", mode & S_IWUSR);
    setboolean(L, "execute", mode & S_IXUSR);
    if (mode & S_IRUSR) bits[1] = 'r';
    if (mode & S_IWUSR) bits[2] = 'w';
    if (mode & S_IXUSR) bits[3] = 'x';
#ifdef __OS2__
    rc = DosQueryPathInfo((PCSZ)fn, FIL_STANDARD, &fstat, sizeof(fstat));
    if (rc == 0) {
      fs = &fstat;  /* do not change this, otherwise Agena will crash */
      unsigned int attribs = fs->attrFile;
      setboolean(L, "hidden", attribs & FILE_HIDDEN);
      setboolean(L, "readonly", attribs & FILE_READONLY);
      setboolean(L, "system", attribs & FILE_SYSTEM);
      setboolean(L, "archived", attribs & FILE_ARCHIVED);
      setboolean(L, "directory", attribs & FILE_DIRECTORY);
      if (attribs & FILE_DIRECTORY) bits[11] = 'd';
      if (attribs & FILE_READONLY) bits[12] = 'r';
      if (attribs & FILE_HIDDEN) bits[13] = 'h';
      if (attribs & FILE_ARCHIVED) bits[14] = 'a';
      if (attribs & FILE_SYSTEM) bits[15] = 's';
    }
#elif defined(_WIN32)
    winmodet = GetFileAttributes(fn);
    setboolean(L, "hidden", winmodet & FILE_ATTRIBUTE_HIDDEN);
    setboolean(L, "readonly", winmodet & FILE_ATTRIBUTE_READONLY);
    setboolean(L, "system", winmodet & FILE_ATTRIBUTE_SYSTEM);
    setboolean(L, "archived", winmodet & FILE_ATTRIBUTE_ARCHIVE);
    if (winmodet & FILE_ATTRIBUTE_READONLY) bits[12] = 'r';
    if (winmodet & FILE_ATTRIBUTE_HIDDEN) bits[13] = 'h';
    if (winmodet & FILE_ATTRIBUTE_ARCHIVE) bits[14] = 'a';
    if (winmodet & FILE_ATTRIBUTE_SYSTEM) bits[15] = 's';
#endif
    lua_settable(L, -3);
#if (defined(__unix__) && !defined(LUA_DOS)) || defined(__APPLE__) || defined(__HAIKU__)
    /* determine group rights */
    lua_pushstring(L, "group");
    lua_createtable(L, 0, 3);
    setboolean(L, "read", mode & S_IRGRP);
    setboolean(L, "write", mode & S_IWGRP);
    setboolean(L, "execute", mode & S_IXGRP);
    lua_settable(L, -3);
    if (mode & S_IRGRP) bits[4] = 'r';
    if (mode & S_IWGRP) bits[5] = 'w';
    if (mode & S_IXGRP) bits[6] = 'x';
    /* determine other users' rights */
    lua_pushstring(L, "other");
    lua_createtable(L, 0, 3);
    setboolean(L, "read", mode & S_IROTH);
    setboolean(L, "write", mode & S_IWOTH);
    setboolean(L, "execute", mode & S_IXOTH);
    lua_settable(L, -3);
    if (mode & S_IROTH) bits[7] = 'r';
    if (mode & S_IWOTH) bits[8] = 'w';
    if (mode & S_IXOTH) bits[9] = 'x';
#endif
    /* 0.26.0, push file mode (mode_t) and permission `bits` */
    setnumber(L, "perms", mode);
    setstring(L, "bits", bits);
  } else
    lua_pushfail(L);  /* file does not exist */
  return 1;
}


static int os_exists (lua_State *L) {
  lua_pushboolean(L, (access(luaL_checkstring(L, 1), 00|04) != -1));  /* 00 = file exists, 04 = is readable */
  return 1;
}


/* print working directory; July 07, 2008; modified 0.26.0, 04.08.2009 (PATH_MAX) */
static int os_curdir (lua_State *L) {
  char *buffer = malloc(sizeof(char)*PATH_MAX);
  getcwd(buffer, PATH_MAX);
  if (buffer == NULL)
    lua_pushfail(L);
  else {
    charreplace(buffer, '\\', '/');
    lua_pushstring(L, buffer);
  }
  free(buffer);
  return 1;
}


/* The following routines were taken from the lposix.c file in the
*  POSIX library for Lua 5.0. Based on original by Claudio Terra for Lua 3.x.
*  Luiz Henrique de Figueiredo <lhf@tecgraf.puc-rio.br>
*  05 Nov 2003 22:09:10
*/

/** mkdir(path), modified for Agena; extended in 0.13.3 to support OS/2 */
static int os_mkdir (lua_State *L) {
#ifdef _WIN32
  const char *path = luaL_checkstring(L, 1);
  return os_pushresult(L, mkdir(path), path);
#elif defined(__unix__) || defined(__OS2__) || defined(__APPLE__) || defined(__HAIKU__)
  const char *path = luaL_checkstring(L, 1);
  return os_pushresult(L, mkdir(path, 0777), path);
#else
  lua_pushfail(L);
  return 1;
#endif
}

/* chdir(path) */
static int os_chdir (lua_State *L) {
  const char *path = luaL_checkstring(L, 1);
  return os_pushresult(L, chdir(path), path);
}

/* rmdir(path) */
static int os_rmdir (lua_State *L) {
  const char *path = luaL_checkstring(L, 1);
  return os_pushresult(L, rmdir(path), path);
}


/* beep: sound the loudspeaker with a given frequency (first arg, a posint) and
   a duration (second arg, in seconds, a float), April 01, 2007 */

static int os_beep (lua_State *L) {
  int nargs;
#if defined(__unix__) || defined(__APPLE__) || defined(__HAIKU__)
  int i;
#endif
  lua_Number duration;
  nargs = lua_gettop(L);
  duration = 1;
#ifdef _WIN32
  if (nargs == 0) {
    putc('\007', stderr);
    lua_pushnil(L);
  } else {
    int tone;
    tone = agn_checknumber(L, 1);
    duration = agn_checknumber(L, 2);
    if (tone > 0 && duration > 0) {
      Beep(tone, duration*1000);
      lua_pushnil(L);
    } else
      lua_pushfail(L);
  }
#elif defined(__OS2__)
  if (nargs == 0) {
    putc('\007', stderr);
    lua_pushnil(L);
  } else {
    ULONG tone;
    tone = agn_checknumber(L, 1);
    duration = agn_checknumber(L, 2);
    if (tone > 0 && duration > 0) {
      if (DosBeep(tone, (ULONG)(duration*1000)))  /* invalid frequency */
        lua_pushfail(L);
      else
        lua_pushnil(L);
    } else
      lua_pushfail(L);
  }
#elif defined(__unix__) || defined(__APPLE__) || defined(__HAIKU__)
  duration = 1;
  if (nargs == 2) {
    duration = (int)agn_checknumber(L, 2);
  }
  for (i=0; i < duration; i++)
    putc('\007', stderr);
  lua_pushnil(L);
#else
  lua_pushfail(L);
#endif
  return 1;
}


/* wait: wait for x seconds; March 31, 2007 */

static int os_wait (lua_State *L) {
  lua_Number x = agn_checknumber(L, 1);
  if (x > 0) {
    #ifdef _WIN32
    Sleep(x*1000);
    lua_pushnil(L);
    #elif defined(__unix__) || defined(__OS2__) || defined(__APPLE__) || defined(__HAIKU__)
    sleep(x);
    lua_pushnil(L);
    #else
    lua_pushfail(L);
    #endif
  } else lua_pushfail(L);
  return 1;
}


/* retrieve information on your platform;
0.12.2, October 12, 2008; extended to correctly identify Vista, Windows 2003, and Windows 2007 Server
on April 08, 2009 - 0.13.4 */

#ifdef _WIN32

typedef struct WinVer {
  int BuildNumber;
  int MajorVersion;
  int MinorVersion;
  int PlatformId;
  char ProductType;
} WinVer;


int getWindowsVersion (struct WinVer *winver) {
   int result, MajorVersion, MinorVersion, ProductType, PlatformId;
   OSVERSIONINFOEX version;
   version.dwOSVersionInfoSize = sizeof(version);
   GetVersionEx((OSVERSIONINFO *)&version);

   /* dwMajorVersion
   Major version number of the operating system. This member can be one of the following values. Value Meaning
   4 Windows NT 4.0, Windows Me, Windows 98, or Windows 95.
   5 Windows Server 2003, Windows XP, or Windows 2000
   6 Windows Vista, Windows Server 2008, Windows 7

   dwMinorVersion
   Minor version number of the operating system. This member can be one of the following values. Value Meaning
   0 Windows 2000, Windows NT 4.0, Windows 95, or Windows Vista
   1 Windows XP, Windows 2008 server
   2 Windows Server 2003, Windows 7
   10 Windows 98
   90 Windows Me

   dwPlatformId
   Operating system platform. This member can be one of the following values. Value Meaning
   VER_PLATFORM_WIN32_NT
   2 Windows Server 2003, Windows XP, Windows 2000, or Windows NT.
   VER_PLATFORM_WIN32_WINDOWS
   1 Windows Me, Windows 98, or Windows 95.
   VER_PLATFORM_WIN32s
   0 Win32s on Windows 3.1. */

   winver->BuildNumber = (int)version.dwBuildNumber;
   MajorVersion = winver->MajorVersion = (int)version.dwMajorVersion;
   MinorVersion = winver->MinorVersion = (int)version.dwMinorVersion;
   PlatformId = winver->PlatformId = (int)version.dwPlatformId;
   ProductType = winver->ProductType = (char)version.wProductType;
   if (MajorVersion == 4 && MinorVersion == 0 && PlatformId == 1)
      result = 2;  /* Win 95 */
   else if (MajorVersion == 4 && MinorVersion == 0 && PlatformId == 2)
      result = 3;  /* Win NT 4.0 */
   else if (MajorVersion == 4 && MinorVersion == 10)
      result = 4;  /* Win 98 */
   else if (MajorVersion == 5 && MinorVersion == 0)
      result = 5;  /* Win 2000 */
   else if (MajorVersion == 4 && MinorVersion == 90)
      result = 6;  /* Win Me */
   else if (MajorVersion == 5 && MinorVersion == 1)
      result = 7;  /* Win XP */
   else if (MajorVersion == 6 && MinorVersion == 0 && ProductType == VER_NT_WORKSTATION)
      result = 8;  /* Win Vista */
   else if (MajorVersion == 6 && MinorVersion >= 0 && ProductType != VER_NT_WORKSTATION)
      result = 9;  /* Win Server 2008 */
   else if (MajorVersion == 5 && MinorVersion == 2)
      result = 10; /* Win 2003 */
   else if (MajorVersion == 6 && MinorVersion == 1 && ProductType == VER_NT_WORKSTATION)
      result = 11; /* Windows 7 */
   else if (PlatformId == 0)
      result = 1;  /* Win32s under Windows 3.1 */
   else
      result = 0;
   return result;
}
#endif


/* Novell DOS 7 = 'IBMPcDos', '6', '00', 'pc'
   OS/2 Warp 4 Fixpack 15 = 'OS/2', '1', '2.45', 'i386'
   Windows = 'Windows', etc. */

static int os_system (lua_State *L) {
#ifdef _WIN32
   /* dwMajorVersion
   Major version number of the operating system. This member can be one of the following values. Value Meaning
   4 Windows NT 4.0, Windows Me, Windows 98, or Windows 95.
   5 Windows Server 2003, Windows XP, or Windows 2000
   6 Windows Vista, Windows Server 2008, Windows 7

   dwMinorVersion
   Minor version number of the operating system. This member can be one of the following values. Value Meaning
   0 Windows 2000, Windows NT 4.0, Windows 95, or Windows Vista
   1 Windows XP, Windows 2008 server
   2 Windows Server 2003, Windows 7
   10 Windows 98
   90 Windows Me

   dwPlatformId
   Operating system platform. This member can be one of the following values. Value Meaning
   VER_PLATFORM_WIN32_NT
   2 Windows Server 2003, Windows XP, Windows 2000, or Windows NT.
   VER_PLATFORM_WIN32_WINDOWS
   1 Windows Me, Windows 98, or Windows 95.
   VER_PLATFORM_WIN32s
   0 Win32s on Windows 3.1. */

  int winver;
  struct WinVer winversion;
  static char *versions[] = {  /* extended 0.13.4 */
    "unknown", "32s", "95", "NT 4.0", "98", "2000", "ME", "XP", "Vista", "Server 2008", "Server 2003", "7"
  };
  winver = getWindowsVersion(&winversion);
  lua_newtable(L);
  lua_rawsetistring(L, -1, 1, "Windows");
  lua_rawsetistring(L, -1, 2, versions[winver]);
  lua_rawsetinumber(L, -1, 3, winversion.BuildNumber);
  lua_rawsetinumber(L, -1, 4, winversion.PlatformId);
  lua_rawsetinumber(L, -1, 5, winversion.MajorVersion);  /* added 0.13.4 */
  lua_rawsetinumber(L, -1, 6, winversion.MinorVersion);  /* added 0.13.4 */
  lua_rawsetinumber(L, -1, 7, winversion.ProductType);   /* added 0.13.4 */
#elif defined(__unix__) || defined(__OS2__) || defined(__APPLE__) || defined(__HAIKU__)
  /* applies to DJGPP compiled DOS code as well; on Apple returns Darwin
  version number */
  struct utsname info;
  if (uname(&info) != -1) {
    lua_newtable(L);
    lua_rawsetistring(L, -1, 1, info.sysname);
    lua_rawsetistring(L, -1, 2, info.release);
    lua_rawsetistring(L, -1, 3, info.version);
    lua_rawsetistring(L, -1, 4, info.machine);
   }
   else
     lua_pushfail(L);
#else
  lua_pushfail(L);
#endif
  return 1;
}


static int os_endian (lua_State *L) {
  int r = tools_endian();
  if (r == -1)
    lua_pushfail(L);
  else
    lua_pushinteger(L, r);
  return 1;
}


static int os_isANSI (lua_State *L) {  /* 0.28.2 */
#ifdef LUA_ANSI
  lua_pushtrue(L);
#else
  lua_pushfalse(L);
#endif
  return 1;
}


/* return status of battery on laptops, works in Windows 2000 and higher. 0.14.0, March 13, 2009 */
static int os_battery (lua_State *L) {
#ifdef _WIN32
  BYTE bACLineStatus, status, bBatteryLifePercent, bReserved1;
  DWORD bBatteryLifeTime, bBatteryFullLifeTime;
  SYSTEM_POWER_STATUS battstat;
  int BatteryCharging, winver;
  struct WinVer windowsver;
  static char *ACstatus[] = {"off", "on", "unknown"};
  static char *BatteryStatusMapping[] = {"medium", "high", "low", "critical", "charging", "no battery", "unknown"};
  winver = getWindowsVersion(&windowsver);
  if (winver < 2 || winver == 3) {  /* Windows NT 4.0 and earlier are not supported */
    lua_pushfail(L);
    return 1;
  }
  /* get battery information by calling the Win32 API */
  GetSystemPowerStatus(&battstat);
  bReserved1 = battstat.Reserved1;
  bACLineStatus = battstat.ACLineStatus;
  BatteryCharging = 0; /* not charging */
  if (bACLineStatus == 255)
    bACLineStatus = 2;
  switch (battstat.BatteryFlag) {
    case 0: /* >= 33 and <= 66 % */
      status = 0;
      break;
    case 1: /* high */
      status = 1;
      break;
    case 2: /* low */
      status = 2;
      break;
    case 4: /* critical */
      status = 3;
      break;
    case 8: case 9: case 10:  /* charging */
      status = battstat.BatteryFlag - 8;
      BatteryCharging = 1;
      break;
    case 12:  /* charging & critical */
      status = 3;
      BatteryCharging = 1;
    case 128: /* No system battery */
      status = 5;
      break;
    case 255: /* unknown status */
      status = 6;
      break;
    default:  /* got value 0 at around 45% on Acer, 10 right after reconnecting AC line  */
      status = 6;
      break;
  }
  bBatteryLifePercent = battstat.BatteryLifePercent;
  bBatteryLifeTime = battstat.BatteryLifeTime;
  bBatteryFullLifeTime = battstat.BatteryFullLifeTime;
  lua_createtable(L, 0, 0);
  setstring(L, "acline", ACstatus[bACLineStatus]);
  if (status == 4) {
    setboolean(L, "installed", 0);
  } else {
    setboolean(L, "installed", 1);
    setnumber(L, "life", bBatteryLifePercent);
    setstring(L, "status", BatteryStatusMapping[status]);
    setboolean(L, "charging", (BatteryCharging == 1));  /* 0.31.4 fix */
    setnumber(L, "flag", battstat.BatteryFlag);
    lua_pushstring(L, "lifetime");
    if (bBatteryLifeTime != -1)
      lua_pushnumber(L, bBatteryLifeTime);
    else
      lua_pushundefined(L);
    lua_settable(L, -3);
    lua_pushstring(L, "fulllifetime");
    if (bBatteryFullLifeTime != -1)
      lua_pushnumber(L, bBatteryFullLifeTime);
    else
      lua_pushundefined(L);
    lua_settable(L, -3);
  }
#elif __OS2__
#define IOCTL_POWER 0x000c
#define POWER_GETPOWERSTATUS 0x0060
  int result;
  HFILE pAPI;
  ULONG call;
  struct _Battery {
    USHORT usParamLen;
    USHORT usPowerFlags;
    UCHAR  ucACStatus;
    UCHAR  ucBatteryStatus;
    UCHAR  ucBatteryLife;
  } Battery;
  struct _Data {
    USHORT usReturn;
  } Data;
  ULONG sizeBattery, sizeData;
  result = DosOpen((PCSZ)"\\DEV\\APM$", &pAPI, &call, 0,
    FILE_NORMAL, FILE_OPEN, OPEN_ACCESS_READONLY | OPEN_SHARE_DENYNONE,
    NULL);
  if (result != 0) {
    lua_pushfail(L);
    return 1;
  }
  sizeBattery = Battery.usParamLen = sizeof(Battery);
  sizeData = sizeof(Data);
  result = DosDevIOCtl(pAPI,
    IOCTL_POWER, POWER_GETPOWERSTATUS, (PVOID)&Battery, sizeof(Battery),
    &sizeBattery, (PVOID)&Data, sizeof(Data), &sizeData);
  if (result == ERROR_INVALID_PARAMETER) {
    lua_pushfail(L);
  } else {
    int acstatus, batterystatus;
    static char *ACstatus[] = {"off", "on", "unknown", "invalid"};
    static char *BatteryStatusMapping[] =
      {"high", "low", "critical", "charging", "unknown", "invalid"};
    lua_newtable(L);
    switch (Battery.ucACStatus) {
      case   0: acstatus = 0; break;
      case   1: acstatus = 1; break;
      case 255: acstatus = 2; break;
      default:  acstatus = 3; break;
    }
    switch (Battery.ucBatteryStatus) {
      case 0:   batterystatus = 0; break;
      case 1:   batterystatus = 1; break;
      case 2:   batterystatus = 2; break;
      case 3:   batterystatus = 3; break;
      case 255: batterystatus = 4; break;
      default:  batterystatus = 5; break;
    }
    setstring(L, "acline", ACstatus[acstatus]);
    setstring(L, "status", BatteryStatusMapping[batterystatus]);
    setnumber(L, "flags", (lua_Number)Battery.usPowerFlags);
    setboolean(L, "powermanagement", ((Battery.usPowerFlags & 0x0001) == 1));
    lua_pushstring(L, "life");
    if (Battery.ucACStatus != 1)  /* AC line is off */
      lua_pushnumber(L, (lua_Number)Battery.ucBatteryLife);
    else
      lua_pushundefined(L);
    lua_rawset(L, -3);
  }
#else
  lua_pushfail(L);
#endif
  return 1;
}


/* 0.25.0, July 19, 2009; extended 26.07.2009, 0.25.1; patched 0.26.0, August 06, 2009 */
static int os_fcopy (lua_State *L) {
  const char *fin, *fout;
  int in, out, bufsize, result;
  size_t bytes;
  mode_t mode;
  struct stat entry;
  result = 0;
  fin = agn_checkstring(L, 1);
  fout = agn_checkstring(L, 2);
  agn_getenv(L, "BufferSize");  /* 0.32.0 */
  bufsize = ( lua_type(L, -1) == LUA_TNUMBER ) ? agn_tonumber(L, -1) : BUFSIZ;
  agn_poptop(L);  /* pop BufferSize */
  if (bufsize < 1) bufsize = BUFSIZ;
  char buffer[bufsize];
  if ( (in = my_roopen(fin)) == -1) {
    lua_pushfail(L);
    return 1;
  }
  if ( (out = my_create(fout)) == -1) {
    close(in);
    lua_pushfail(L);
    return 1;
  }
  while ( (bytes = read(in, buffer, bufsize)) )
    write(out, buffer, bytes);
  close(in);
  close(out);
  if (bytes < 0)
    lua_pushfail(L);
  else {
    if (stat(fin, &entry) == 0) {
      /* set file permission attributes */
      mode = entry.st_mode;
#if (defined(__unix__) && !defined(LUA_DOS)) || defined(__APPLE__) || defined(__HAIKU__)
      result = chmod(fout, (mode & S_IRUSR)|(mode & S_IWUSR)|(mode & S_IXUSR)|(mode & S_IRGRP)|(mode & S_IWGRP)|(mode & S_IXGRP)|(mode & S_IROTH)|(mode & S_IWOTH)|(mode & S_IXOTH));
#elif defined(_WIN32)
      /* added 0.26.0, 07.08.2009 */
      result = !(SetFileAttributes(fout, GetFileAttributes(fin)));
#elif defined(__OS2__)
      static FILESTATUS3 fstat, *fs;
      APIRET rc = DosQueryPathInfo((PCSZ)fin, FIL_STANDARD, &fstat, sizeof(fstat));
      if (rc == 0) {
        fs = &fstat;  /* do not change this, otherwise Agena will crash */
        DosSetPathInfo((PCSZ)fout, FIL_STANDARD, fs, sizeof(FILESTATUS3), 0);
      }
#else
      result = chmod(fout, (mode & S_IRUSR)|(mode & S_IWUSR)|(mode & S_IXUSR));
#endif
      if (result == 0)
        lua_pushtrue(L);
      else
        lua_pushfail(L);
    }
    else
      lua_pushfail(L);
  }
  return 1;
}


static int os_drives (lua_State *L) {
#if defined(_WIN32) || defined(__OS2__)
  unsigned int i, c;
  int anydrivefound;
  char drive[1];
  #ifdef _WIN32
  long buff, mask;
  buff = GetLogicalDrives();
  #else
  ULONG buff, mask, curDrive;
  APIRET rc;
  rc = DosQueryCurrentDisk(&curDrive, &buff);
  if (rc != 0)
    luaL_error(L, "error determining drives.");
  #endif
  anydrivefound = c = i = 0;
  agn_createseq(L, 26);
  while( i < 26 ) {
    mask = (1 << i);
    if( (buff & mask) > 0 ) {
      drive[0] = 65+i;
      lua_pushlstring(L, drive, 1);
      lua_seqseti(L, -2, ++c);
      /* prevent error messages if no media are inserted in floppy drives or CD-ROMs */
      anydrivefound = 1;
    }
    i++;
  }
  if (!anydrivefound) {
    agn_poptop(L);  /* delete sequence */
    lua_pushfail(L);
  }
#else
  lua_pushfail(L);
#endif
  return 1;
}


static int os_drivestat (lua_State *L) {
#ifdef _WIN32
  unsigned int drivetype;
  size_t l;
  long buff, mask;
  char drive[4];
  DWORD serial_number;
  DWORD max_component_length;
  DWORD file_system_flags;
  char name [256];
  char file_system_name[256];
  unsigned long sectors_per_cluster, bytes_per_sector, number_of_free_clusters, total_number_of_clusters;
  double freespace, totalspace;
  static char *types[] = {
    "unknown", "no root dir", "Removable", "Fixed", "Remote", "CD-ROM", "RAMDISK"};
  const char *drivename = agn_checklstring(L, 1, &l);
  if (l < 1) {  /* empty string ? */
    lua_pushfail(L);
    return 1;
  }
  drive[0] = toupper(*drivename); drive[1] = ':'; drive[2] = '\\'; drive[3] = '\0';
  buff = GetLogicalDrives();
  mask = (1 << (drive[0]-65));
  if( (buff & mask) < 1 ) {  /* no valid drive ? */
    lua_pushfail(L);
    return 1;
  }
  /* set defaults to prevent invalid values when calling GetDiskFreeSpace */
  sectors_per_cluster = 0, bytes_per_sector = 0, number_of_free_clusters = 0, total_number_of_clusters = 0;
  /* set defaults to prevent invalid values when calling GetVolumeInformation  */
  strcpy(name, ""); strcpy(file_system_name, "");
  serial_number = 0; max_component_length = 0; file_system_flags = 0;
  drivetype = GetDriveType(drive);
  lua_createtable(L, 0, 5);
  GetVolumeInformation(drive, name,
    (sizeof (name)), &serial_number, &max_component_length, &file_system_flags,
    file_system_name, (sizeof (file_system_name)));
  if (name == NULL || strcmp(name, "") == 0)
    strcpy(name, "<none>");
  if (strlen(name) > 19) name[19] = '\0';
  if (file_system_name == NULL || strcmp(file_system_name, "") == 0)
    strcpy(file_system_name, "<unknown>");
  if (strlen(file_system_name) > 11) file_system_name[11] = '\0';

  /* prevent error messages if no media are inserted in floppy drives or CD-ROMs */
  SetErrorMode(0x0001);

  /* get info on free space on drive processed */
  GetDiskFreeSpace(drive, &sectors_per_cluster, &bytes_per_sector,
    &number_of_free_clusters, &total_number_of_clusters);

  SetErrorMode(0); /* set error mode back to normal (any exceptions printed again) */

  freespace = (double)number_of_free_clusters * (double)sectors_per_cluster * (double)bytes_per_sector;
  totalspace = (double)total_number_of_clusters * (double)sectors_per_cluster * (double)bytes_per_sector;

  setstring(L, "label", name);
  setstring(L, "drivetype", (drivetype < 7) ? types[drivetype] : "unknown");
  setstring(L, "filesystem",
    (strcmp(file_system_name, "") == 0) ? "<unknown>" : file_system_name);
  setnumber(L, "freesize", freespace);
  setnumber(L, "totalsize", totalspace);
#else
  lua_pushfail(L);
#endif
  return 1;
}


static mode_t newmode (mode_t oldmode, mode_t modeflag, int add) {
  mode_t modet = oldmode;
  if (add)
    modet |= modeflag;
  else if (modet & modeflag)
    modet -= modeflag;
  return modet;
}

#ifdef _WIN32
static DWORD winmode (DWORD oldmode, DWORD modeflag, int add) {
  DWORD winmodet = oldmode;
  if (add)
    winmodet |= modeflag;
  else if (winmodet & modeflag)
    winmodet -= modeflag;
  return winmodet;
}
#elif defined(__OS2__)
static unsigned int os2mode (unsigned int oldmode, unsigned int modeflag, int add) {
  unsigned int os2modet = oldmode;
  if (add)
    os2modet |= modeflag;
  else if (os2modet & modeflag)
    os2modet -= modeflag;
  return os2modet;
}
#endif

static int os_fattrib (lua_State *L) {  /* 0.26.0 */
  const char *fn, *mode;
  mode_t modet;
  struct stat entry;
  int result, owner, add;
  size_t l, i;
#if (defined(__unix__) && !defined(LUA_DOS)) || defined(__APPLE__) || defined(__HAIKU__)
  size_t j;
#elif defined(_WIN32)
  DWORD winmodet;
#elif defined(__OS2__)
  static FILESTATUS3 fstat, *fs;
  unsigned int os2modet;
  APIRET rc;
#endif
  modet = owner = add = 0;
  fn = agn_checkstring(L, 1);
#if defined(_WIN32)
  winmodet = GetFileAttributes(fn);
#elif defined(__OS2__)
  rc = DosQueryPathInfo((PCSZ)fn, FIL_STANDARD, &fstat, sizeof(fstat));
  if (rc != 0) luaL_error(L, "could not determine file attributes.");
  fs = &fstat;
  os2modet = fs->attrFile;
#endif
  if (stat(fn, &entry) == 0)
    modet = entry.st_mode;
  else
    luaL_error(L, "file does not exist.");
  mode = agn_checklstring(L, 2, &l);
  if (l < 3)
    luaL_error(L, "malformed settings string.");
#if !defined(__unix__) && !defined(__APPLE__) && !defined(__HAIKU__)
  mode_t modew[1] = {S_IWUSR};
  mode_t modex[1] = {S_IXUSR};
#else
#ifndef LUA_DOS
  mode_t moder[3] = {S_IRUSR, S_IRGRP, S_IROTH};
#else
  mode_t moder[1] = {S_IRUSR};
#endif
  mode_t modew[3] = {S_IWUSR, S_IWGRP, S_IWOTH};
  mode_t modex[3] = {S_IXUSR, S_IXGRP, S_IXOTH};
#endif
  switch (*mode++) {
    case 'u': case 'U': {
      owner = 0;  /* owner */
      break;
    }
    case 'g': case 'G': {
      owner = 1;  /* group */
      break;
    }
    case 'o': case 'O': {
      owner = 2;  /* other */
      break;
    }
    case 'a': case 'A': {
      owner = 3;  /* all */
      break;
    }
    default:
      luaL_error(L, "wrong user specification, must be `u` , `g`, `o`, or `a`.");
  }
#if ((!defined(__unix__) && !defined(__APPLE__) &&  !defined(__HAIKU__)) || defined(LUA_DOS))
  if (owner > 0) owner = 0;
#endif
  switch (*mode++) {
    case '+': {
      add = 1;  /* owner */
      break;
    }
    case '-': {
      add = 0;  /* group */
      break;
    }
    default:
      luaL_error(L, "wrong add/delete specification, must be `+` , or `-`.");
  }
  for (i=2; i < l; i++, mode++) {
    switch(*mode) {
#if defined(__unix__) || defined(__APPLE__) || defined(__HAIKU__)
      case 'r': case 'R': {
        if (owner < 3)
          modet = newmode(modet, moder[owner], add);
#ifndef LUA_DOS
        else
          for (j=0; j < 3; j++) {modet = newmode(modet, moder[j], add);}
#endif
        break;
      }
#endif
      case 'w': case 'W': {
        if (owner < 3)
          modet = newmode(modet, modew[owner], add);
#if (defined(__unix__) && !defined(LUA_DOS)) || defined(__APPLE__) || defined(__HAIKU__)
        else
          for (j=0; j < 3; j++) {modet = newmode(modet, modew[j], add);}
#endif
        break;
      }
      case 'x': case 'X': {
        if (owner < 3)
          modet = newmode(modet, modex[owner], add);
#if (defined(__unix__) && !defined(LUA_DOS)) || defined(__APPLE__) || defined(__HAIKU__)
        else
          for (j=0; j < 3; j++) {modet = newmode(modet, modex[j], add);}
#endif
        break;
      }
#ifdef _WIN32
      case 'h': case 'H': {
        winmodet = winmode(winmodet, FILE_ATTRIBUTE_HIDDEN, add);
        break;
      }
      case 'a': case 'A': {
        winmodet = winmode(winmodet, FILE_ATTRIBUTE_ARCHIVE, add);
        break;
      }
      case 's': case 'S': {
        winmodet = winmode(winmodet, FILE_ATTRIBUTE_SYSTEM, add);
        break;
      }
      case 'r': case 'R': {
        winmodet = winmode(winmodet, FILE_ATTRIBUTE_READONLY, add);
        break;
      }
#elif defined(__OS2__)
      case 'h': case 'H': {
        os2modet = os2mode(os2modet, FILE_HIDDEN, add);
        break;
      }
      case 'a': case 'A': {
        os2modet = os2mode(os2modet, FILE_ARCHIVED, add);
        break;
      }
      case 's': case 'S': {
        os2modet = os2mode(os2modet, FILE_SYSTEM, add);
        break;
      }
      case 'r': case 'R': {
        os2modet = os2mode(os2modet, FILE_READONLY, add);
        break;
      }
#endif
      default: lua_assert(0);
    }
  }
  result = chmod(fn, modet);
#ifdef _WIN32
  result = result || !(SetFileAttributes(fn, winmodet));
#elif defined(__OS2__)
  fs->attrFile = os2modet;
  rc = DosSetPathInfo((PCSZ)fn, FIL_STANDARD, fs, sizeof(FILESTATUS3), 0);
  result = result || rc;
#endif
  if (result == 0)
     lua_pushtrue(L);
  else
     lua_pushfail(L);
  return 1;
}


static int os_screensize (lua_State *L) {  /* 0.31.3 */
#ifdef __WIN32
  lua_pushnumber(L, GetSystemMetrics(SM_CXSCREEN));
  lua_pushnumber(L, GetSystemMetrics(SM_CYSCREEN));
  agn_createpair(L, -2, -1);
  lua_remove(L, -2);
  lua_remove(L, -2);
#else
  lua_pushfail(L);
#endif
  return 1;
}


static int os_mousebuttons (lua_State *L) {  /* 0.31.3 */
#ifdef __WIN32
  lua_pushnumber(L, GetSystemMetrics(SM_CMOUSEBUTTONS));
#else
  lua_pushfail(L);
#endif
  return 1;
}


static const luaL_Reg syslib[] = {
  {"battery",   os_battery},          /* 0.14.0, added on March 13, 2009 */
  {"beep",      os_beep},             /* added on March 31, 2007 */
  {"chdir",     os_chdir},            /* added July 07, 2008 */
  {"fattrib",   os_fattrib},          /* 0.26.0, added August 06, 2009 */
  {"computername", os_computername},  /* 0.5.3  Sept 15, 2007 */
  {"curdir",    os_curdir},           /* added July 07, 2008 */
  {"date",      os_date},
  {"difftime",  os_difftime},
  {"drives",    os_drives},           /* 0.26.0, 05.08.2009 */
  {"drivestat", os_drivestat},        /* 0.26.0, 05.08.2009 */
  {"endian",    os_endian},           /* 0.8.0, added on December 11, 2007 */
  {"execute",   os_execute},          /* changed in 0.5.4 */
  {"exists",    os_exists},           /* added on April 09, 2007 */
  {"exit",      os_exit},
  {"fcopy",     os_fcopy},            /* added on July 19, 2009 */
  {"freemem",   os_freemem},          /* 0.5.3  Sept 15, 2007 */
  {"fstat",     os_fstat},            /* added on October 11, 2008 */
  {"getenv",    os_getenv},
  {"isANSI",    os_isANSI},           /* added 0.28.2, 17.11.2009 */
  {"listcore",  os_listcore},         /* added on August 01, 2007 */
  {"login",     os_login},            /* 0.5.3  Sept 15, 2007 */
  {"memstate",  os_memstate},         /* 0.5.3  Sept 15, 2007 */
  {"mkdir",     os_mkdir},            /* added July 07, 2008 */
  {"mousebuttons", os_mousebuttons},  /* added March 28, 2010 */
  {"move",      os_move},
  {"remove",    os_remove},
  {"rmdir",     os_rmdir},            /* added July 07, 2008 */
  {"screensize", os_screensize},      /* added March 28, 2010 */
  {"setlocale", os_setlocale},
  {"system",    os_system},           /* 0.12.2  October 12, 2008 */
  {"time",      os_time},
  {"tmpname",   os_tmpname},
  {"wait",      os_wait},             /* added on March 31, 2007 */
  {NULL, NULL}
};

/* }====================================================== */



LUALIB_API int luaopen_os (lua_State *L) {
  luaL_register(L, LUA_OSLIBNAME, syslib);
  return 1;
}


