/*
** $Id: utils.c, v 1.0.1 by Alexander Walz - Initiated: October 17, 2006
   Revision 08.08.2009 22:13:15
** See Copyright Notice in agena.h
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

/* the following package ini declarations must be included after `#include <` and before `include #` ! */

#define utils_c
#define LUA_LIB

#include "agena.h"

#include "agnxlib.h"
#include "agenalib.h"

#include "utils.h"  /* typedefs and checkarray macro */

/* macro to `unsign' a character */
#define uchar(c)        ((unsigned char)(c))

/*
----------------------------------------------------------------------------
  General string manipulation and small search functions
----------------------------------------------------------------------------
*/

static int str_newarray (lua_State *L) {
  int i, n;
  size_t nbytes;
  stringarray *a;
  n = luaL_checkint(L, 1);
  luaL_argcheck(L, n>=1, 1, "invalid size");
  nbytes = sizeof(stringarray) + n * (sizeof(const char *) + sizeof(size_t));
  a = (stringarray *)lua_newuserdata(L, nbytes);
  a->size = n;
  for (i=0; i<n; i++) {
    a->data[i].entry = NULL;
    a->data[i].len = 0;
  }
  luaL_getmetatable(L, "stringarray");
  lua_setmetatable(L, -2);
  return 1;
}


static int str_arraysize (lua_State *L) {
  stringarray *a = checkarray(L, 1);
  lua_pushinteger(L, a->size);
  return 1;
}


static int str_setarray (lua_State *L) {
  size_t l;
  stringarray *a = checkarray(L, 1);
  int index = luaL_checkint(L, 2);
  const char *str = agn_checklstring(L, 3, &l);
  luaL_argcheck(L, 0 <= index && index < a->size, 2, "index out of range");
  a->data[index].entry = strdup(str);
  a->data[index].len = l;
  return 0;
}


static int str_getarray (lua_State *L) {
  stringarray *a = checkarray(L, 1);
  int index = luaL_checkint(L, 2);
  luaL_argcheck(L, 0 <= index && index < a->size, 2, "index out of range");
  lua_pushlstring(L, a->data[index].entry, a->data[index].len);
  return 1;
}


static int str_getwholearray (lua_State *L) {
  int i;
  stringarray *a = checkarray(L, 1);
  lua_createtable(L, a->size, 0);
  for (i=0; i<a->size; i++) {
    if (a->data[i].entry != NULL)
      lua_rawsetilstring(L, -1, i+1, a->data[i].entry, a->data[i].len);
  }
  return 1;
}


static int str_searcharray (lua_State *L) {
  int i;
  stringarray *a = checkarray(L, 1);
  const char *str = agn_checkstring(L, 2);
  for (i=0; i<a->size; i++) {
    if (a->data[i].entry != NULL && strcmp(a->data[i].entry, str) == 0) {
      lua_pushinteger(L, i+1);
      return 1;
    }
  }
  lua_pushnil(L);
  return 1;
}


static int str_array2string (lua_State *L) {
  stringarray *a = checkarray(L, 1);
  lua_pushfstring(L, "array(%d)", a->size);
  return 1;
}


static int str_arraygc (lua_State *L) {
  (void)L;
  return 0;
}


static const struct luaL_Reg str_arraylib [] = {
  {"__writeindex", str_setarray},
  {"__index", str_getarray},
  {"__tostring", str_array2string},
  {"__gc", str_arraygc},
  {NULL, NULL}
};


/* substitute individual characters in str by replacements in array, patched 0.12.2 */

static int str_singlesubs (lua_State *L) {
  int i;
  size_t l;
  unsigned char token;
  const char *src = agn_checklstring(L, 1, &l);
  stringarray *a = checkarray(L, 2);
  luaL_Buffer b;
  luaL_buffinit(L, &b);
  for (i=0; i<l; i++) {
    token = uchar(src[i]);
    /* token < a->size: check whether value of char is less than overall array size to avoid crashes */
    if (token < a->size && a->data[token].entry != NULL)
      luaL_addlstring(&b, a->data[token].entry, a->data[token].len);
    else
      luaL_addchar(&b, token);
  }
  luaL_pushresult(&b);
  return 1;
}

/*
----------------------------------------------------------------------------
  Internal utilities
----------------------------------------------------------------------------
*/

static void setfield (lua_State *L, const char *key, int value) {
  lua_pushinteger(L, value);
  lua_setfield(L, -2, key);
}


static void setboolfield (lua_State *L, const char *key, int value) {
  if (value < 0)  /* undefined? */
    return;  /* does not set field */
  lua_pushboolean(L, value);
  lua_setfield(L, -2, key);
}

/*
----------------------------------------------------------------------------
  Other functions
----------------------------------------------------------------------------
*/

static int sys_calendar (lua_State *L) {
  time_t seconds = luaL_opt(L, (time_t)agn_checknumber, 1, time(NULL));
  struct tm *stm;
  lua_createtable(L, 0, 9);  /* 9 = number of fields */
  if (seconds < 0) {  /* 0.31.3 patch */
    lua_pushfail(L);
    return 1;
  }
  stm = localtime(&seconds);
  if (stm != NULL) {
    setfield(L, "sec", stm->tm_sec);
    setfield(L, "min", stm->tm_min);
    setfield(L, "hour", stm->tm_hour);
    setfield(L, "day", stm->tm_mday);
    setfield(L, "month", stm->tm_mon+1);
    setfield(L, "year", stm->tm_year+1900);
    setfield(L, "wday", stm->tm_wday+1);
    setfield(L, "yday", stm->tm_yday+1);
    setboolfield(L, "DST", stm->tm_isdst);
  } else
    lua_pushfail(L);
  return 1;
}


static const luaL_Reg utils[] = {
  {"singlesubs", str_singlesubs},         /* added on August 05, 2007 */
  {"calendar", sys_calendar},             /* added on April 07, 2007 */
  {"newarray", str_newarray},             /* added on August 05, 2007 */
  {"arraysize", str_arraysize},           /* added on August 05, 2007 */
  {"setarray", str_setarray},             /* added on August 05, 2007 */
  {"getarray", str_getarray},             /* added on August 05, 2007 */
  {"searcharray", str_searcharray},       /* added on August 12, 2007 */
  {"getwholearray", str_getwholearray},   /* added on August 08, 2007 */
  {NULL, NULL}
};


static void createmetatable (lua_State *L) {
  lua_createtable(L, 0, 1);  /* create metatable for strings */
  lua_pushliteral(L, "");  /* dummy string */
  lua_pushvalue(L, -2);
  lua_setmetatable(L, -2);  /* set string metatable */
  agn_poptop(L);  /* pop dummy string */
  lua_pushvalue(L, -2);  /* string library... */
  lua_setfield(L, -2, "__index");  /* ...is the __index metamethod */
  agn_poptop(L);  /* pop metatable */
}


/*
** Open utils library
*/
LUALIB_API int luaopen_utils (lua_State *L) {
  luaL_newmetatable(L, "stringarray");
  luaL_register(L, NULL, str_arraylib);
  luaL_register(L, AGENA_UTILSLIBNAME, utils);
  createmetatable(L);
  return 1;
}

/* ====================================================================== */

